#! /usr/bin/env python
#
# Test(s) for ../ipython_utils.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_ipython_utils.py
#

"""Tests for ipython_utils module"""

# Standard packages
## TODO: from collections import defaultdict

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper
from mezcla import debug
import mezcla.ipython_utils as THE_MODULE

#------------------------------------------------------------------------

class TestIt(TestWrapper):
    """Class for command-line based testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    def test_grep_obj_methods(self):
        """Test (module) object grepping"""
        debug.trace(4, f"TestIt.test_grep_obj_methods(); self={self}")
        import_functions = THE_MODULE.grep_obj_methods(THE_MODULE, "import_*")
        assert ("import_module_globals" in import_functions)
        leading_space_names = THE_MODULE.grep_obj_methods(THE_MODULE, "^ +")
        assert (not leading_space_names)
        return

    def test_import_module_globals(self):
        """Test module globals import"""
        debug.trace(4, f"TestIt.test_import_module_globals(); self={self}")
        #
        globals_dict = {}
        THE_MODULE.import_module_globals(self.script_module, globals_dict=globals_dict)
        assert ("TL" in globals_dict)
        #
        globals_dict = {}
        THE_MODULE.import_module_globals(THE_MODULE, globals_dict=globals_dict, ignore_errors=True)
        assert (not globals_dict)
        return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# Wrapper class around unittest.TestCase
#
# Notes:
# - Based on template.py used in older test scripts
# - Creates per-test temp file, based on same class-wide temp-base file.
# - To treat temp-base as a subdirectory, set use_temp_base_dir to True in 
#   class member initialiation section.
# - Changes to temporary directory/file should be synchronized with ../main.py.
# - Overriding the temporary directory can be handy during debugging; however,
#   you might need to specify different ones if you invoke helper scripts. See
#   tests/test_extract_company_info.py for an example.
# TODO:
# - * Clarify TEMP_BASE vs. TEMP_FILE usage.
# - Clarify that this can co-exist with pytest-based tests (see tests/test_main.py).
# TODO2:
# - Clean up script_file usage (and unncessary settings in test scripts).
#
#-------------------------------------------------------------------------------
# Sample test (streamlined version of test_simple_main_example.py):
#
#   import unittest
#   from mezcla import system
#   from mezcla.unittest_wrapper import TestWrapper
#
#   class TestIt(TestWrapper):
#       """Class for testcase definition"""
#       script_module = TestWrapper.derive_tested_module_name(__file__)
#  
#       def test_simple_data(self):
#           """Make sure simple data sample processed OK"""
#           system.write_file(self.temp_file, "really fubar")
#           output = self.run_script("--check-fubar", self.temp_file)
#           self.assertTrue("really" in output)
#
#   if __name__ == '__main__':
#      unittest.main()
#-------------------------------------------------------------------------------
# TODO:
# - Add method to invoke unittest.main(), so clients don't need to import unittest.
# - Clarify how ALLOW_SUBCOMMAND_TRACING affects tests that invoke external scripts.
#

"""Unit test support class"""

# Standard packages
import inspect
import os
import tempfile
import unittest

# Installed packages
## TODO: import pytest

# Local packages
import mezcla
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system

# Constants (e.g., environment options)

TL = debug.TL
KEEP_TEMP = system.getenv_bool("KEEP_TEMP", debug.detailed_debugging(),
                               "keep temporary files")
TODO_FILE = "TODO FILE"
TODO_MODULE = "TODO MODULE"

# Note: the following is for transparent resolution of dotted module names
# for invocation of scripts via 'python -m package.module'. This is in support
# of transitioning from the old way of importing packages via 'import module'
# instead of 'import package.module'. (The former required that package be
# explicitly specified in the python path, such as via 'PYTHONPATH=package-dir:...'.)
THIS_PACKAGE = getattr(mezcla.debug, "__package__", None)
debug.assertion(THIS_PACKAGE == "mezcla")

def get_temp_dir(keep=False):
    """Get temporary directory, omitting later deletion if KEEP"""
    if keep is None:
        keep = KEEP_TEMP
    dir_path = tempfile.NamedTemporaryFile(delete=(not keep)).name
    gh.full_mkdir(dir_path)
    debug.trace(5, "get_temp_dir() => {dir_path}")
    return dir_path


def trap_exception(function):
    """Decorator to trap exception during function execution
    Note:
    - Only intended for use in tests (e.g., fix for maldito pytest).
    - Issues assertion so that test fails.
    - Should be inside any pytest.mark.xfail decorators.
    """
    debug.trace(8, f"trap_exception({gh.elide(function)}")
    #
    def wrapper(*args):
        """Wrapper around variable arity function f"""   ## TODO: {function.__name__}
        debug.trace(7, f"in wrapper: args={args}")
        result = None
        try:
            result = function(*args)
        except AssertionError:
            raise
        except:
            system.print_exception_info(function)
            assert(False)
        return result
    #
    debug.trace(7, f"trap_exception() => {gh.elide(wrapper)}")
    return wrapper


def pytest_fixture_wrapper(function):
    """Decorator for use with pytest fixtures like capsys
    Usage:
        @pytest_fixture_wrapper
        @trap_exception
        def test_it(capsys):
            ...
    """
    # See https://stackoverflow.com/questions/19614658/how-do-i-make-pytest-fixtures-work-with-decorated-functions    
    debug.trace(8, f"pytest_fixture_wrapper({gh.elide(function)}")
    #
    def wrapper(x):
        """Wrapper around unary function f(x)"""   ## TODO: {function.__name__}
        debug.trace(7, f"in wrapper: x={x}")
        return function(x)
    #
    debug.trace(7, f"pytest_fixture_wrapper() => {gh.elide(wrapper)}")
    return wrapper


class TestWrapper(unittest.TestCase):
    """Class for testcase definition"""
    script_file = TODO_FILE             # path for invocation via 'python -m coverage run ...' (n.b., usually set via get_module_file_path)
    script_module = TODO_MODULE         # name for invocation via 'python -m' (n.b., usually set via derive_tested_module_name)
    temp_base = system.getenv_text("TEMP_BASE",
                                   tempfile.NamedTemporaryFile().name)
    check_coverage = system.getenv_bool("CHECK_COVERAGE", False,
                                        "Check coverage during unit testing")
    ## TODO: temp_file = None
    ## TEMP: initialize to unique value independent of temp_base
    temp_file = None
    use_temp_base_dir = system.is_directory(temp_base)
    test_num = 1
    
    ## TEST:
    ## NOTE: leads to pytest warning. See
    ##   https://stackoverflow.com/questions/62460557/cannot-collect-test-class-testmain-because-it-has-a-init-constructor-from
    ## def __init__(self, *args, **kwargs):
    ##     debug.trace_fmtd(5, "TestWrapper.__init__({a}): keywords={kw}; self={s}",
    ##                      a=",".join(args), kw=kwargs, s=self)
    ##     super().__init__(*args, **kwargs)
    ##    debug.trace_object(5, self, label="TestWrapper instance")
    ##
    ## __test__ = False                 # make sure not assumed test
        
    @classmethod
    def setUpClass(cls):
        """Per-class initialization: make sure script_module set properly"""
        debug.trace_fmtd(5, "TestWrapper.setupClass(): cls={c}", c=cls)
        super().setUpClass()
        debug.trace_object(5, cls, "TestWrapper class")
        debug.assertion(cls.script_module != TODO_MODULE)
        if cls.script_module:
            # Try to pull up usage via python -m mezcla.xyz --help
            help_usage = gh.run("python -m '{mod}' --help", mod=cls.script_module)
            debug.assertion("No module named" not in help_usage,
                            f"problem running via 'python -m {cls.script_module}'")
            # Warn about lack of usage statement unless "not intended for command-line" type warning issued
            # TODO: standardize the not-intended wording
            if (not ((my_re.search(r"warning:.*not intended", help_usage,
                                   flags=my_re.IGNORECASE))
                     or ("usage:" in help_usage.lower()))):
                system.print_stderr("Warning: script should implement --help")

        # Optionally, setup temp-base directory (normally just a file)
        if cls.use_temp_base_dir is None:
            cls.use_temp_base_dir = system.getenv_bool("USE_TEMP_BASE_DIR", False)
            # TODO: temp_base_dir = system.getenv_text("TEMP_BASE_DIR", " "); cls.use_temp_base_dir = bool(temp_base_dir.strip()); ...
        if cls.use_temp_base_dir:
            ## TODO: pure python
            ## TODO: gh.full_mkdir
            gh.run("mkdir -p {dir}", dir=cls.temp_base)

        return

    @staticmethod
    def derive_tested_module_name(test_filename):
        """Derive the name of the module being tested from TEST_FILENAME. Used as follows:
              script_module = TestWrapper.derive_tested_module_name(__file__)
        Note: *** Deprecated method *** (see get_testing_module_name)
        """
        debug.trace(3, "Warning: in deprecrated derive_tested_module_name")
        module = os.path.split(test_filename)[-1]
        module = my_re.sub(r".py[oc]?$", "", module)
        module = my_re.sub(r"^test_", "", module)
        debug.trace_fmtd(5, "derive_tested_module_name({f}) => {m}",
                         f=test_filename, m=module)
        return (module)

    @staticmethod
    def get_testing_module_name(test_filename, module_object=None):
        """Derive the name of the module being tested from TEST_FILENAME and MODULE_OBJECT
        Note: used as follows (see tests/test_template.py):
            script_module = TestWrapper.get_testing_module_name(__file__)
        """
        # Note: Used to resolve module name given THE_MODULE (see template).
        module_name = os.path.split(test_filename)[-1]
        module_name = my_re.sub(r".py[oc]?$", "", module_name)
        module_name = my_re.sub(r"^test_", "", module_name)
        package_name = THIS_PACKAGE
        if module_object is not None:
           package_name = getattr(module_object, "__package__", "")
           debug.trace_expr(4, package_name)
        if package_name:
            full_module_name = package_name + "." + module_name
        else:
            full_module_name = module_name
        debug.trace_fmtd(4, "get_testing_module_name({f}, [{mo}]) => {m}",
                         f=test_filename, m=full_module_name, mo=module_object)
        return (full_module_name)

    @staticmethod
    def get_module_file_path(test_filename):
        """Return absolute path of module being tested"""
        result = system.absolute_path(test_filename)
        result = my_re.sub(r'tests\/test_(.*\.py)', r'\1', result)
        debug.assertion(result.endswith(".py"))
        debug.trace(7, f'get_module_file_path({test_filename}) => {result}')
        return result

    def setUp(self):
        """Per-test initializations
        Notes:
        - Disables tracing scripts invoked via run() unless ALLOW_SUBCOMMAND_TRACING
        - Initializes temp file name (With override from environment)."""
        # Note: By default, each test gets its own temp file.
        debug.trace(4, "TestWrapper.setUp()")
        if not gh.ALLOW_SUBCOMMAND_TRACING:
            gh.disable_subcommand_tracing()
        # The temp file is an extension of temp-base file by default.
        # Optionally, if can be a file in temp-base subdrectory.
        if self.use_temp_base_dir:
            default_temp_file = gh.form_path(self.temp_base, "test-")
        else:
            default_temp_file = self.temp_base + "-test-"
        default_temp_file += str(TestWrapper.test_num)
        self.temp_file = system.getenv_text("TEMP_FILE", default_temp_file)
        gh.delete_existing_file(self.temp_file)
        TestWrapper.test_num += 1

        debug.trace_object(5, self, "TestWrapper instance")
        return

    def run_script(self, options=None, data_file=None, log_file=None, trace_level=4,
                   out_file=None, env_options=None, uses_stdin=None, post_options=None, background=None):
        """Runs the script over the DATA_FILE (optional), passing (positional)
        OPTIONS and optional setting ENV_OPTIONS. If OUT_FILE and LOG_FILE are
        not specifed, they  are derived from self.temp_file. The optional POST_OPTIONS
        go after the data file.
        Notes:
        - issues warning if script invocation leads to error
        - if USES_STDIN, requires explicit empty string for DATA_FILE to avoid use of - (n.b., as a precaution against hangups)"""
        debug.trace_fmtd(trace_level + 1,
                         "TestWrapper.run_script(opts={opts}, data={df}, log={lf}, lvl={lvl}, out={of}, env={env}, stdin={stdin}, post={post}, back={back})",
                         opts=options, df=data_file, lf=log_file, lvl=trace_level, of=out_file,
                         env=env_options, stdin=uses_stdin, post=post_options, back=background)
        if options is None:
            options = ""
        if env_options is None:
            env_options = ""
        if post_options is None:
            post_options = ""

        # Derive the full paths for data file and log, and then invoke script.
        # TODO: derive from temp base and data file name?;
        # TODO1: derive default for uses_stdin based on use of filename argment (e.g., from usage)
        uses_stdin_false = ((uses_stdin is not None) and not bool(uses_stdin))
        data_path = ("" if uses_stdin_false else "-")
        if data_file is not None:
            data_path = (gh.resolve_path(data_file) if len(data_file) else data_file)
        if not log_file:
            log_file = self.temp_file + ".log"
        if not out_file:
            out_file = self.temp_file + ".out"
        # note: output is redirected to a file to preserve tabs

        # Set converage script path and command spec
        coverage_spec = ''
        if self.check_coverage:
            debug.assertion(self.script_file)
            self.script_module = self.script_file
            coverage_spec = 'coverage run'
        else:
            debug.assertion(not self.script_module.endswith(".py"))
        amp_spec = "&" if background else ""

        # Run the command
        gh.issue("{env} python -m {cov_spec} {module}  {opts}  {path}  {post} 1> {out} 2> {log} {amp_spec}",
                 env=env_options, cov_spec=coverage_spec, module=self.script_module,
                 opts=options, path=data_path, out=out_file, log=log_file, post=post_options, amp_spec=amp_spec)
        output = system.read_file(out_file)
        # note; trailing newline removed as with shell output
        if output.endswith("\n"):
            output = output[:-1]
        debug.trace_fmtd(trace_level, "output: {{\n{out}\n}}",
                         out=gh.indent_lines(output))

        # Make sure no python or bash errors. For example,
        #   "SyntaxError: invalid syntax" and "bash: python: command not found"
        log_contents = system.read_file(log_file)
        error_found = my_re.search(r"(\S+error:)|(no module)|(command not found)",
                                   log_contents.lower())
        debug.assertion(not error_found)
        debug.trace_expr(trace_level + 1, log_contents, max_len=2048)

        # Do sanity check for python exceptions
        traceback_found = my_re.search("Traceback.*most recent call", log_contents)
        debug.assertion(not traceback_found)

        return output

    def do_assert(self, condition, message=None):
        """Shows context for assertion failure with CONDITION and then issue assert
        If MESSAGE specified, included in assertion error
        Note:
        - Works around for maldito pytest, which makes it hard to do simple things like pinpointing errors.
        - Formatted similar to debug.assertion:
             Test assertion failed: <expr> (at <f><n>): <msg>
        """
        if ((not condition) and debug.debugging(debug.TL.DEFAULT)):
            statement = filename = line_num = None
            try:
                # note: accounts for trap_exception and other decorators
                for caller in inspect.stack():
                    debug.trace_expr(8, caller)
                    (_frame, filename, line_num, _function, context, _index) = caller
                    statement = debug.read_line(filename, line_num).strip()
                    if "do_assert" in statement:
                        break
                debug.trace_expr(7, filename, line_num, context, prefix="do_assert: ")
            except:
                system.print_exception_info("do_assert")
            debug.assertion(statement)
            if statement:
                # TODO3: use abstract syntax tree (AST) based extraction
                # ex: self.do_assert(not my_re.search(r"cat|dog", description))  # no pets
                # Isolate condition
                cond = my_re.sub(r"^\s*\S+\.do_assert\((.*)\)", r"\1", statement)
                # Get expression proper, removing optional comments and semicolon 
                expr = my_re.sub(r";?\s*#.*$", "", cond)
                # Strip optional message
                qual = ""
                if message is not None:
                    expr = my_re.sub(r", *([\'\"]).*\1\s*$", "", expr)   # string arg
                    expr = my_re.sub(r", *[a-z0-9_]+$", "", expr,        # variable arg
                                     flags=my_re.IGNORECASE)
                    qual = f": {message}"
                # Format assertion error with optional qualification (i.e., user message)
                debug.trace(1, f"Test assertion failed: {expr} (at {filename}:{line_num}){qual}")
                debug.trace(5, f"\t{statement}")
            else:
                system.print_error("Warning: unexpected condition in do_assert")
        assert(condition)
    
    def tearDown(self):
        """Per-test cleanup: deletes temp file unless detailed debugging"""
        debug.trace(4, "TestWrapper.tearDown()")
        if not KEEP_TEMP:
            gh.run("rm -vf {file}*", file=self.temp_file)
        return

    @classmethod
    def tearDownClass(cls):
        """Per-class cleanup: stub for tracing purposes"""
        debug.trace_fmtd(5, "TestWrapper.tearDownClass(); cls={c}", c=cls)
        if not KEEP_TEMP:
            ## TODO: use shutil
            if cls.use_temp_base_dir:
                gh.run("rm -rvf {dir}", dir=cls.temp_base)
            else:
                gh.run("rm -vf {base}*", base=cls.temp_base)
        super().tearDownClass()
        return
    
#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_DETAILED)
    debug.trace(TL.USUAL, "Warning: not intended for command-line use\n")



#! /usr/bin/env python
#
# Test(s) for ../system.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_system.py
#

"""Tests for system module"""

# Standard packages
import io
from math import pi
import time
import sys
import re

# Installed packages
import pytest
import pickle

# Local packages
from mezcla import glue_helpers as gh
from mezcla import debug

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.system as THE_MODULE

class TestSystem:
    """Class for test case definitions"""

    def test_maxint(self):
        """Ensure maxint works as expected"""
        debug.trace(4, "test_maxint()")
        ## TODO: WORK-IN=PROGRESS

    def test_register_env_option(self):
        """Ensure register_env_option works as expected"""
        debug.trace(4, "test_register_env_option()")

        THE_MODULE.env_options = {}
        THE_MODULE.env_defaults = {}

        THE_MODULE.register_env_option(
            var='VAR_STRING',
            description='this is a string variable',
            default='empty'
        )

        assert THE_MODULE.env_options['VAR_STRING'], 'this is a string variable'
        assert THE_MODULE.env_defaults['VAR_STRING'], 'empty'
        assert len(THE_MODULE.env_options) == 1
        assert len(THE_MODULE.env_defaults) == 1

    def test_get_registered_env_options(self):
        """Ensure get_registered_env_options works as expected"""
        debug.trace(4, "test_get_registered_env_options()")

        set_test_env_var()

        assert isinstance(THE_MODULE.get_registered_env_options(), list)
        assert 'VAR_STRING' in THE_MODULE.get_registered_env_options()
        assert len(THE_MODULE.get_registered_env_options()) == 2

    def test_get_environment_option_descriptions(self):
        """Ensure get_environment_option_descriptions works as expected"""
        debug.trace(4, "test_get_environment_option_descriptions()")

        set_test_env_var()

        ## TODO: Test include_all option
        ## assert ('VAR_STRING', "this is a string variable") in THE_MODULE.get_environment_option_descriptions(include_all=False)

        # Test include_default option
        assert ('VAR_STRING', "this is a string variable") in THE_MODULE.get_environment_option_descriptions(include_default=False)
        assert ('VAR_STRING', "this is a string variable ('empty')") in THE_MODULE.get_environment_option_descriptions(include_default=True)

        # Test indent option
        assert ('VAR_STRING', "this is a string variable, default=('empty')") in THE_MODULE.get_environment_option_descriptions(indent=', default=')
        assert ('VAR_STRING', "this is a string variable => ('empty')") in THE_MODULE.get_environment_option_descriptions(indent=' => ')

        # Test get N enviroments vars descriptions
        assert len(THE_MODULE.get_environment_option_descriptions()) == 2

    def test_formatted_environment_option_descriptions(self):
        """Ensure formatted_environment_option_descriptions works as expected"""
        debug.trace(4, "test_formatted_environment_option_descriptions()")

        set_test_env_var()

        # Test sort
        expected = (
            'VAR_STRING\tthis is a string variable (\'empty\')\n'
            '\tANOTHER_VAR\tthis is another env. var. (\'2022\')'
        )
        assert THE_MODULE.formatted_environment_option_descriptions(sort=False) == expected
        expected = (
            'ANOTHER_VAR\tthis is another env. var. (\'2022\')\n'
            '\tVAR_STRING\tthis is a string variable (\'empty\')'
        )
        assert THE_MODULE.formatted_environment_option_descriptions(sort=True) == expected

        # Test include_all
        # NOTE: this is being tested on test_system.test_get_environment_option_descriptions()

        # Test indent
        expected = (
            'VAR_STRING + this is a string variable (\'empty\')\n'
            ' + ANOTHER_VAR + this is another env. var. (\'2022\')'
        )
        assert THE_MODULE.formatted_environment_option_descriptions(indent=' + ') == expected

    def test_getenv(self, monkeypatch):
        """Ensure getenv works as expected"""
        debug.trace(4, "test_getenv()")
        monkeypatch.setenv('TEST_ENV_VAR', 'some value', prepend=False)
        assert THE_MODULE.getenv('TEST_ENV_VAR') == 'some value'
        assert THE_MODULE.getenv('INT_ENV_VAR', default_value=5) == 5

    def test_getenv_text(self, monkeypatch):
        """Ensure getenv_text works as expected"""
        debug.trace(4, "test_getenv_text()")
        monkeypatch.setenv('TEST_ENV_VAR', 'some value', prepend=False)
        assert THE_MODULE.getenv_text('TEST_ENV_VAR') == 'some value'
        assert not THE_MODULE.getenv_text("REALLY FUBAR?", False)

    def test_getenv_value(self, monkeypatch):
        """Ensure getenv_value works as expected"""
        debug.trace(4, "test_getenv_value()")
        set_test_env_var()
        monkeypatch.setenv('NEW_ENV_VAR', 'some value', prepend=False)
        assert THE_MODULE.getenv_value('NEW_ENV_VAR', default='empty', description='another test env var') == 'some value'
        assert THE_MODULE.env_defaults['NEW_ENV_VAR'] == 'empty'
        assert THE_MODULE.env_options['NEW_ENV_VAR'] == 'another test env var'

    def test_getenv_bool(self, monkeypatch):
        """Ensure getenv_bool works as expected"""
        debug.trace(4, "test_getenv_bool()")
        # note: whitespaces is not a typo, is to test strip condition
        monkeypatch.setenv('TEST_BOOL', 'FALSE', prepend=False)
        assert not THE_MODULE.getenv_bool('TEST_BOOL', None)
        monkeypatch.setenv('TEST_BOOL', '  true   ', prepend=False)
        assert THE_MODULE.getenv_bool('TEST_BOOL', None)
        assert isinstance(THE_MODULE.getenv_bool('TEST_BOOL', None), bool)

    def test_getenv_number(self, monkeypatch):
        """Ensure getenv_number works as expected"""
        debug.trace(4, "test_getenv_number()")
        # note: whitespaces is not a typo, is to test strip condition
        monkeypatch.setenv('TEST_NUMBER', ' 9.81    ', prepend=False)
        assert THE_MODULE.getenv_number('TEST_NUMBER', default=10) == 9.81
        assert THE_MODULE.getenv_number('BAD_TEST_NUMBER', default=10) == 10
        ## TODO: test helper argument

    def test_getenv_int(self, monkeypatch):
        """Ensure getenv_int works as expected"""
        debug.trace(4, "test_getenv_int()")
        monkeypatch.setenv('TEST_NUMBER', '9.81', prepend=False)
        assert THE_MODULE.getenv_int('TEST_NUMBER', default=20) == 9
        assert THE_MODULE.getenv_int("REALLY FUBAR", 123) == 123

    def test_get_exception(self):
        """Ensure get_exception works as expected"""
        debug.trace(4, "test_get_exception()")
        ## TODO: WORK-IN=PROGRESS

    def test_print_error(self, capsys):
        """Ensure print_error works as expected"""
        debug.trace(4, "test_print_error()")
        THE_MODULE.print_error("this is an test error message")
        captured = capsys.readouterr()
        assert "error" in captured.err

    def test_print_stderr(self, capsys):
        """Ensure print_stderr works as expected"""
        debug.trace(4, "test_print_stderr()")
        THE_MODULE.print_stderr("Error: F{oo}bar!", oo='OO')
        captured = capsys.readouterr()
        assert "Error: FOObar!" in captured.err

    def test_print_exception_info(self, capsys):
        """Ensure print_exception_info works as expected"""
        debug.trace(4, "test_print_exception_info()")
        THE_MODULE.print_exception_info("Foobar")
        captured = capsys.readouterr()
        assert "Foobar" in captured.err

    @pytest.mark.xfail
    def test_exit(self, monkeypatch, capsys):
        """Ensure exit works as expected"""
        debug.trace(4, "test_exit()")
        def sys_exit_mock():
            return 'exit'
        monkeypatch.setattr(sys, "exit", sys_exit_mock)
        assert THE_MODULE.exit('test exit {method}', method='method') == 'exit'
        # Exit is mocked, ignore code editor hidding
        captured = capsys.readouterr()
        assert "test exit method" in captured.err

    def test_setenv(self):
        """Ensure setenv works as expected"""
        debug.trace(4, "test_setenv()")
        THE_MODULE.setenv('NEW_TEST_ENV_VAR', 'the gravity is 10, pi is 3')
        assert THE_MODULE.getenv('NEW_TEST_ENV_VAR') == 'the gravity is 10, pi is 3'

    def test_print_full_stack(self):
        """Ensure print_full_stack works as expected"""
        debug.trace(4, "test_print_full_stack()")
        ## TODO: WORK-IN=PROGRESS

    def test_trace_stack(self):
        """Ensure trace_stack works as expected"""
        debug.trace(4, "test_trace_stack()")
        ## TODO: WORK-IN=PROGRESS

    def test_get_current_function_name(self):
        """Ensure get_current_function_name works as expected"""
        debug.trace(4, "test_get_current_function_name()")
        assert THE_MODULE.get_current_function_name() == "test_get_current_function_name"

    def test_open_file(self):
        """Ensure open_file works as expected"""
        debug.trace(4, "test_open_file()")
        ## TODO: WORK-IN=PROGRESS

    def test_save_object(self):
        """Ensure save_object works as expected"""
        debug.trace(4, "test_save_object()")
        test_dict = {
            1: 'first',
            2: 'second',
        }
        test_filename = gh.get_temp_file()

        THE_MODULE.save_object(test_filename, test_dict)

        test_file = open(test_filename, 'rb')
        actual_object = pickle.load(test_file)
        assert actual_object == test_dict
        test_file.close()

    def test_load_object(self, capsys):
        """Ensure load_object works as expected"""
        debug.trace(4, "test_load_object()")

        # Test valid file
        test_dict = {
            1: 'first',
            2: 'second',
        }
        test_filename = gh.get_temp_file()
        test_file = open(test_filename, 'wb')
        pickle.dump(test_dict, test_file)
        test_file.close()
        assert THE_MODULE.load_object(test_filename) == test_dict

        # Test invalid file
        THE_MODULE.load_object('bad_file_name')
        captured = capsys.readouterr()
        assert "Error:" in captured.err

    def test_quote_url_text(self):
        """Ensure quote_url_text works as expected"""
        debug.trace(4, "test_quote_url_text()")

        ## TODO: set fixed sys.version_info.major > 2.

        # Test quoting
        assert THE_MODULE.quote_url_text("<2/") == "%3C2%2F"
        assert THE_MODULE.quote_url_text("Joe's hat") == "Joe%27s+hat"
        assert THE_MODULE.quote_url_text("Joe%27s+hat") == "Joe%2527s%2Bhat"

        # Test unquoting
        assert THE_MODULE.quote_url_text("%3C2%2f", unquote=True) == "<2/"
        assert THE_MODULE.quote_url_text("Joe%27s+hat", unquote=True) == "Joe's hat"
        assert THE_MODULE.quote_url_text("Joe%2527s%2Bhat", unquote=True) == "Joe%27s+hat"

        ## TODO: Test sys.version_info.major < 2

    def test_unquote_url_text(self):
        """Ensure unquote_url_text works as expected"""
        debug.trace(4, "test_unquote_url_text()")
        assert THE_MODULE.unquote_url_text("%3C2%2f") == "<2/"
        assert THE_MODULE.unquote_url_text("Joe%27s+hat") == "Joe's hat"
        assert THE_MODULE.unquote_url_text("Joe%2527s%2Bhat") == "Joe%27s+hat"

    def test_escape_html_text(self):
        """Ensure escape_html_text works as expected"""
        debug.trace(4, "test_escape_html_text()")

        ## TODO: set fixed sys.version_info.major > 2.

        assert THE_MODULE.escape_html_text("<2/") == "&lt;2/"
        assert THE_MODULE.escape_html_text("Joe's hat") == "Joe&#x27;s hat"

        ## TODO: test with sys.version_info.major < 2

    def test_unescape_html_text(self):
        """Ensure unescape_html_text works as expected"""
        debug.trace(4, "test_unescape_html_text()")

        ## TODO: set fixed sys.version_info.major > 2.

        assert THE_MODULE.unescape_html_text("&lt;2/") == "<2/"
        assert THE_MODULE.unescape_html_text("Joe&#x27;s hat") == "Joe's hat"

        ## TODO: test with sys.version_info.major < 2

    def test_stdin_reader(self, monkeypatch):
        """Ensure stdin_reader works as expected"""
        debug.trace(4, "test_stdin_reader()")
        monkeypatch.setattr('sys.stdin', io.StringIO('my input\nsome line\n'))
        test_iter = THE_MODULE.stdin_reader()
        assert next(test_iter) == 'my\tinput'
        assert next(test_iter) == 'some\tline'
        assert next(test_iter) == ''

    def test_read_all_stdin(self, monkeypatch):
        """Ensure read_all_stdin works as expected"""
        debug.trace(4, "test_read_all_stdin()")
        monkeypatch.setattr('sys.stdin', io.StringIO('my input\nsome line'))
        assert THE_MODULE.read_all_stdin() == 'my input\nsome line'

    def test_read_entire_file(self, capsys):
        """Ensure read_entire_file works as expected"""
        debug.trace(4, "test_read_entire_file()")

        # Test valid vile
        temp_file = gh.get_temp_file()
        gh.write_file(temp_file, 'file\nwith\nmultiple\nlines\n')
        assert THE_MODULE.read_entire_file(temp_file) == 'file\nwith\nmultiple\nlines\n'

        # Test invalid file
        debug.set_level(3)
        THE_MODULE.read_entire_file('invalid_file', errors='ignore')
        captured = capsys.readouterr()
        assert "Unable to read file" not in captured.err
        THE_MODULE.read_entire_file('invalid_file')
        captured = capsys.readouterr()
        assert "Unable to read file" in captured.err

    def test_read_lines(self):
        """Ensure read_lines works as expected"""
        debug.trace(4, "test_read_lines()")
        temp_file = gh.get_temp_file()
        gh.write_file(temp_file, 'file\nwith\nmultiple\nlines\n')
        assert THE_MODULE.read_lines(temp_file) == ['file', 'with', 'multiple', 'lines']

    def test_read_binary_file(self):
        """Ensure read_binary_file works as expected"""
        debug.trace(4, "test_read_binary_file()")
        ## TODO: WORK-IN=PROGRESS

    def test_read_directory(self):
        """Ensure read_directory works as expected"""
        debug.trace(4, "test_read_directory()")
        ## TODO: WORK-IN=PROGRESS

    def test_get_directory_filenames(self):
        """Ensure get_directory_filenames works as expected"""
        debug.trace(4, "test_get_directory_filenames()")
        assert "/etc/passwd" in THE_MODULE.get_directory_filenames("/etc")
        assert "/boot" not in THE_MODULE.get_directory_filenames("/", just_regular_files=True)

    def test_read_lookup_table(self, capsys):
        """Ensure read_lookup_table works as expected"""
        debug.trace(4, "test_read_lookup_table()")

        content = (
            'COUNTRY -> CAPITAL\n'
            'United States -> Washington D. C.\n'
            'France -> Paris\n'
            'Canada -> Ottawa\n'
        )
        expected_uppercase = {
            'COUNTRY': 'CAPITAL',
            'United States': 'Washington D. C.',
            'France':'Paris',
            'Canada': 'Ottawa'
        }
        expected_lowercase = {
            'country': 'capital',
            'united states': 'washington d. c.',
            'france':'paris',
            'canada': 'ottawa'
        }

        # Test normal usage parameters
        temp_file = gh.get_temp_file()
        gh.write_file(temp_file, content)
        assert THE_MODULE.read_lookup_table(temp_file, skip_header=False, delim=' -> ', retain_case=False) == expected_lowercase
        assert THE_MODULE.read_lookup_table(temp_file, skip_header=False, delim=' -> ', retain_case=True) == expected_uppercase
        assert THE_MODULE.read_lookup_table(temp_file, skip_header=True, retain_case=False)['country'] == ''

        # Tests default delim
        content_with_tabs = content.replace(' -> ', '\t')
        temp_file = gh.get_temp_file()
        gh.write_file(temp_file, content_with_tabs)
        assert THE_MODULE.read_lookup_table(temp_file) == expected_lowercase

        # Test without delim
        without_delim_content = (
            'line without delim\n'
            'France -> Paris\n'
        )
        temp_file = gh.get_temp_file()
        gh.write_file(temp_file, without_delim_content)
        THE_MODULE.read_lookup_table(temp_file)
        captured = capsys.readouterr()
        assert 'Warning: Ignoring line' in captured.err

        # Test invalid filename
        THE_MODULE.read_lookup_table('bad_filename')
        captured = capsys.readouterr()
        assert 'Error' in captured.err

    def test_create_boolean_lookup_table(self, capsys):
        """Ensure create_boolean_lookup_table works as expected"""
        debug.trace(4, "test_create_boolean_lookup_table()")

        content_with_custom_delim = (
            'EmailEntered - someemail@hotmail.com\n'
            'PasswdEntered - 12345\n'
            'IsBusiness - True\n'
        )
        content_tab_delim = content_with_custom_delim.replace(' - ', '\t')
        expected_lowercase = {
            'emailentered': True,
            'passwdentered': True,
            'isbusiness': True
        }
        expected_uppercase = {
            'EmailEntered': True,
            'PasswdEntered': True,
            'IsBusiness': True
        }

        temp_file = gh.get_temp_file()
        gh.write_file(temp_file, content_with_custom_delim)
        assert THE_MODULE.create_boolean_lookup_table(temp_file, delim=' - ', retain_case=False) == expected_lowercase
        assert THE_MODULE.create_boolean_lookup_table(temp_file, delim=' - ', retain_case=True) == expected_uppercase

        # Test default delim tab
        temp_file = gh.get_temp_file()
        gh.write_file(temp_file, content_tab_delim)
        assert THE_MODULE.create_boolean_lookup_table(temp_file, retain_case=True) == expected_uppercase

        # Test invalid file
        THE_MODULE.create_boolean_lookup_table('/tmp/bad_filename')
        captured = capsys.readouterr()
        assert 'Error:' in captured.err

    def test_lookup_entry(self):
        """Ensure lookup_entry works as expected"""
        debug.trace(4, "test_lookup_entry()")
        hash_map = {
            'description': 'this is a TEST',
            'passwdentered': '12345',
            'IsBusiness': 'True',
        }
        assert THE_MODULE.lookup_entry(hash_map, 'PasswdEntered') == '12345'
        assert THE_MODULE.lookup_entry(hash_map, 'description', retain_case=True) == 'this is a TEST'

    def test_write_file(self):
        """Ensure write_file works as expected"""
        debug.trace(4, "test_write_file()")

        # Test normal usage
        filename = gh.get_temp_file()
        THE_MODULE.write_file(filename, "it")
        assert THE_MODULE.read_file(filename) == "it\n"

        # Test skip newline argument
        filename = gh.get_temp_file()
        THE_MODULE.write_file(filename, "it", skip_newline=True)
        assert THE_MODULE.read_file(filename) == "it"
        assert THE_MODULE.read_file(filename) != "it\n"

    def test_write_binary_file(self):
        """Ensure write_binary_file works as expected"""
        debug.trace(4, "test_write_binary_file()")
        ## TODO: WORK-IN=PROGRESS

    def test_write_lines(self):
        """Ensure write_lines works as expected"""
        debug.trace(4, "test_write_lines()")

        content = (
            'this is\n'
            'a multiline\n'
            'text used\n'
        )
        content_in_lines = [
            'this is',
            'a multiline',
            'text used',
        ]

        # Test normal usage
        filename = gh.get_temp_file()
        THE_MODULE.write_lines(filename, content_in_lines)
        assert THE_MODULE.read_file(filename) == content

        # Test append
        THE_MODULE.write_lines(filename, ['for testing'], append=True)
        assert THE_MODULE.read_file(filename) == content + 'for testing\n'

    def test_write_temp_file(self):
        """Ensure write_temp_file works as expected"""
        debug.trace(4, "test_write_temp_file()")
        temp_file = gh.get_temp_file()
        THE_MODULE.write_temp_file(temp_file, 'some content')
        assert THE_MODULE.read_file(temp_file) == 'some content\n'

    def test_get_file_modification_time(self):
        """Ensure get_file_modification_time works as expected"""
        debug.trace(4, "test_get_file_modification_time()")
        ## TODO: WORK-IN=PROGRESS

    def test_remove_extension(self):
        """Ensure remove_extension works as expected"""
        debug.trace(4, "test_remove_extension()")
        assert THE_MODULE.remove_extension("/tmp/document.pdf") == "/tmp/document"
        assert THE_MODULE.remove_extension("it.abc.def") == "it.abc"
        assert THE_MODULE.remove_extension("it.abc.def", "abc.def") == "it"
        assert THE_MODULE.remove_extension("it.abc.def", ".abc.def") == "it"
        assert THE_MODULE.remove_extension(".coveragerc") == ".coveragerc"

        ## TODO: Check if this is a valid test case and if the returned value is correct
        assert THE_MODULE.remove_extension("it.abc.def", "abc.dtf") == "it.abc.def"

    def test_file_exists(self):
        """Ensure file_exists works as expected"""
        debug.trace(4, "test_file_exists()")
        existent_file = gh.get_temp_file()
        gh.write_file(existent_file, 'content')
        assert THE_MODULE.file_exists(existent_file)
        assert not THE_MODULE.file_exists('bad_file_name')

    def test_get_file_size(self):
        """Ensure get_file_size works as expected"""
        debug.trace(4, "test_get_file_size()")
        temp_file = gh.get_temp_file()
        gh.write_file(temp_file, 'content')
        assert THE_MODULE.get_file_size(temp_file) == 8
        assert THE_MODULE.get_file_size('non-existent-file.txt') == -1

    def test_form_path(self):
        """Ensure form_path works as expected"""
        debug.trace(4, "test_form_path()")
        ## TODO: WORK-IN=PROGRESS

    def test_is_directory(self):
        """Ensure is_directory works as expected"""
        debug.trace(4, "test_is_directory()")
        assert THE_MODULE.is_directory("/etc")

    def test_is_regular_file(self):
        """Ensure is_regular_file works as expected"""
        debug.trace(4, "test_is_regular_file()")
        filename = gh.get_temp_file()
        gh.write_file(filename, 'content')
        assert THE_MODULE.is_regular_file(filename)
        assert not THE_MODULE.is_regular_file('/etc')

    def test_create_directory(self):
        """Ensure create_directory works as expected"""
        debug.trace(4, "test_create_directory()")
        ## TODO: WORK-IN=PROGRESS

    @pytest.mark.xfail
    def test_get_current_directory(self):
        """Ensure get_current_directory works as expected"""
        debug.trace(4, "test_get_current_directory()")
        assert '/home/' in THE_MODULE.get_current_directory()

    def test_set_current_directory(self):
        """Ensure set_current_directory works as expected"""
        debug.trace(4, "test_set_current_directory()")
        ## TODO: WORK-IN=PROGRESS

    def test_to_utf8(self):
        """Ensure to_utf8 works as expected"""
        debug.trace(4, "test_to_utf8()")
        assert THE_MODULE.to_utf8("\ufeff") == "\ufeff"
        ## TODO: add tests for sys.version_info.major < 3
        ## assert THE_MODULE.to_utf8("\ufeff") == "\xEF\xBB\xBF"

    def test_to_str(self):
        """Ensure to_str works as expected"""
        debug.trace(4, "test_to_str()")
        assert THE_MODULE.to_str(pi).startswith('3.141592653589793')

    def test_from_utf8(self):
        """Ensure from_utf8 works as expected"""
        debug.trace(4, "test_from_utf8()")
        ## TODO: WORK-IN=PROGRESS

    def test_to_unicode(self):
        """Ensure to_unicode works as expected"""
        debug.trace(4, "test_to_unicode()")
        assert THE_MODULE.to_unicode("\xEF\xBB\xBF") == "\xEF\xBB\xBF"
        ## TODO: add tests for sys.version_info.major < 3
        ## assert THE_MODULE.to_unicode("\xEF\xBB\xBF") == "\ufeff"

    def test_from_unicode(self):
        """Ensure from_unicode works as expected"""
        debug.trace(4, "test_from_unicode()")
        ## TODO: WORK-IN=PROGRESS

    def test_to_string(self):
        """Ensure to_string works as expected"""
        debug.trace(4, "test_to_string()")
        assert THE_MODULE.to_string(123) == "123"
        assert THE_MODULE.to_string("\u1234") ==  "\u1234"
        assert THE_MODULE.to_string(None) == "None"

    def test_chomp(self):
        """Ensure chomp works as expected"""
        debug.trace(4, "test_chomp()")
        assert THE_MODULE.chomp("some\n") == "some"
        assert THE_MODULE.chomp("abc\n\n") == "abc\n"
        assert THE_MODULE.chomp("http://localhost/", "/") == "http://localhost"

    def test_normalize_dir(self):
        """Ensure normalize_dir works as expected"""
        debug.trace(4, "test_normalize_dir()")
        assert THE_MODULE.normalize_dir("/etc/") == "/etc"

    def test_non_empty_file(self):
        """Ensure non_empty_file works as expected"""
        debug.trace(4, "test_non_empty_file()")

        # Test valid file
        file_with_content = gh.get_temp_file()
        gh.write_file(file_with_content, 'content')
        assert THE_MODULE.non_empty_file(file_with_content)

        # Test non existent file
        assert not THE_MODULE.non_empty_file('bad_file_name')

        # Test empty file
        empty_file = gh.get_temp_file()
        with open(empty_file, 'wb') as _:
            pass # gh.write_file cant be used because appends a newline
        assert not THE_MODULE.non_empty_file(empty_file)

    def test_absolute_path(self):
        """Ensure absolute_path works as expected"""
        debug.trace(4, "test_absolute_path()")
        assert THE_MODULE.absolute_path("/etc/mtab").startswith("/etc")

    @pytest.mark.xfail
    def test_real_path(self):
        """Ensure real_path works as expected"""
        debug.trace(4, "test_real_path()")
        assert THE_MODULE.real_path("/etc/mtab").startswith("/proc")

    def test_get_module_version(self):
        """Ensure get_module_version works as expected"""
        debug.trace(4, "test_get_module_version()")
        ## TODO: WORK-IN=PROGRESS

    def test_intersection(self):
        """Ensure intersection works as expected"""
        debug.trace(4, "test_intersection()")
        assert THE_MODULE.intersection([1, 2], [5, 7, 8]) == []
        assert THE_MODULE.intersection([1, 2, 3, 4, 5], [2, 4]) == [2, 4]
        assert THE_MODULE.intersection([1, 2], [5, 7, 8], as_set=True) == set()
        assert THE_MODULE.intersection([1, 2, 3, 4, 5], [2, 4], as_set=True) == {2, 4}

    def test_union(self):
        """Ensure union works as expected"""
        debug.trace(4, "test_union()")
        assert THE_MODULE.union([1, 2, 3], [2, 3, 4, 5]) == [1, 2, 3, 4, 5]
        assert THE_MODULE.union([1, 2, 3], [2, 3, 4, 5], as_set=True) == {1, 2, 3, 4, 5}

    def test_difference(self):
        """Ensure difference works as expected"""
        debug.trace(4, "test_difference()")
        assert THE_MODULE.difference([5, 4, 3, 2, 1], [4, 2]) == [5, 3, 1]
        assert THE_MODULE.difference([1, 2, 3], [2]) == [1, 3]
        assert THE_MODULE.difference([1, 1, 2, 2], [1]) == [2, 2]

    def test_append_new(self):
        """Ensure append_new works as expected"""
        debug.trace(4, "test_append_new()")
        assert THE_MODULE.append_new([1, 2], 3) == [1, 2, 3]
        assert THE_MODULE.append_new([1, 2, 3], 3) == [1, 2, 3]

    def test_just_one_true(self):
        """Ensure just_one_true works as expected"""
        debug.trace(4, "test_just_one_true()")

        assert THE_MODULE.just_one_true(['yes', '', '', ''], strict=True)
        assert not THE_MODULE.just_one_true(['', '', '', ''], strict=True)

        # all nones should return always True unless strict
        assert THE_MODULE.just_one_true(['', '', '', ''], strict=False)
        assert not THE_MODULE.just_one_true(['', '', '', ''], strict=True)

    def test_just_one_non_null(self):
        """Ensure just_one_non_null works as expected"""
        debug.trace(4, "test_just_one_non_null()")

        assert THE_MODULE.just_one_non_null(['yes', None, None, None], strict=True)
        assert not THE_MODULE.just_one_non_null([None, None, None, None], strict=True)

        # all nones should return always True unless strict
        assert THE_MODULE.just_one_non_null([None, None, None, None], strict=False)
        assert not THE_MODULE.just_one_non_null([None, None, None, None], strict=True)

    def test_unique_items(self):
        """Ensure unique_items works as expected"""
        debug.trace(4, "test_unique_items()")

        # Test normal usage
        assert THE_MODULE.unique_items([1, 2, 3, 2, 1], prune_empty=False) == [1, 2, 3]

        # Test prune_empty
        assert THE_MODULE.unique_items(['cars', 'cars', 'plane', 'train', ''], prune_empty=False) == ['cars', 'plane', 'train', '']
        assert THE_MODULE.unique_items(['cars', 'cars', 'plane', 'train', ''], prune_empty=True) == ['cars', 'plane', 'train']

    def test_to_float(self):
        """Ensure to_float works as expected"""
        debug.trace(4, "test_to_float()")

        # Test normal usage
        assert THE_MODULE.to_float('9.81', default_value=10.0) == 9.81

        # Test default_value
        assert THE_MODULE.to_float('3,14', default_value=4.0) == 4.0 # comma raises the exception.

    def test_to_int(self):
        """Ensure to_int works as expected"""
        debug.trace(4, "test_to_int()")

        # Test normal usage
        assert THE_MODULE.to_int('45', default_value=5, base=10) == 45

        # Test default_value
        assert THE_MODULE.to_int('foo', default_value=5, base=10) == 5

        # Test base
        assert THE_MODULE.to_int('10011', default_value=5, base=2) == 19

    def test_to_bool(self):
        """Ensure to_bool works as expected"""
        debug.trace(4, "test_to_bool()")
        assert THE_MODULE.to_bool(True)
        assert THE_MODULE.to_bool("True")
        assert THE_MODULE.to_bool("Foobar")
        assert THE_MODULE.to_bool(333)
        assert not THE_MODULE.to_bool(False)
        assert not THE_MODULE.to_bool("false")
        assert not THE_MODULE.to_bool("none")
        assert not THE_MODULE.to_bool("off")
        assert not THE_MODULE.to_bool("0")
        assert not THE_MODULE.to_bool("")

    def test_round_num(self):
        """Ensure round_num works as expected"""
        debug.trace(4, "test_round_num()")
        assert THE_MODULE.round_num(3.15914, 3) == 3.159
        THE_MODULE.PRECISION = 5
        assert THE_MODULE.round_num(3.15914311312) == 3.15914

    def test_round_as_str(self):
        """Ensure round_as_str works as expected"""
        debug.trace(4, "test_round_as_str()")
        assert THE_MODULE.round_as_str(3.15914, 3) == "3.159"
        assert isinstance(THE_MODULE.round_as_str(3.15914, 3), str)

    def test_sleep(self, monkeypatch, capsys):
        """Ensure sleep works as expected"""
        debug.trace(4, "test_sleep()")
        def sleep_mock(secs):
            return f'sleeping {secs}'
        monkeypatch.setattr(time, "sleep", sleep_mock)
        THE_MODULE.sleep(123123, trace_level=-1)
        captured = capsys.readouterr()
        assert '123123' in captured.err

    def test_current_time(self, monkeypatch):
        """Ensure current_time works as expected"""
        debug.trace(4, "test_current_time()")
        def time_mock():
            return 12345.6789
        monkeypatch.setattr(time, "time", time_mock)
        assert THE_MODULE.current_time() == 12345.6789
        assert THE_MODULE.current_time(integral=True) == 12346

    def test_time_in_secs(self):
        """Ensure time_in_secs works as expected"""
        debug.trace(4, "test_time_in_secs()")
        # This method is most covered on test_current_time
        assert isinstance(THE_MODULE.time_in_secs(), int)

    def test_python_maj_min_version(self):
        """Ensure python_maj_min_version works as expected"""
        debug.trace(4, "test_python_maj_min_version()")
        assert re.search(r'\d+\.\d+', str(THE_MODULE.python_maj_min_version()))

    def test_get_args(self, monkeypatch):
        """Ensure get_args works as expected"""
        debug.trace(4, "test_get_args()")
        args = [
            "pytest",
            "--name",
            "logfilename.log",
        ]
        monkeypatch.setattr("sys.argv", args)
        assert THE_MODULE.get_args() == args

    def test_main(self, capsys):
        """Ensure main works as expected"""
        THE_MODULE.main('some-arg')
        captured = capsys.readouterr()
        # ex: Warning, tomohara: system.py not intended for direct invocation!
        assert "not intended" in captured.err.lower()


def set_test_env_var():
    """Set enviroment vars to run tests"""
    THE_MODULE.env_options = {
        'VAR_STRING': 'this is a string variable',
        'ANOTHER_VAR': 'this is another env. var.'
    }
    THE_MODULE.env_defaults = {
        'VAR_STRING': 'empty',
        'ANOTHER_VAR': '2022'
    }


if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# gensim_test.py: Tests features of gensim document processing, such ad TF/IDF-based term vector
# representation and documented similarity computations. This takes as input a text file where each
# line represents a separate document.
#
# Notes:
# - gensim is short for 'generate similar' (see http://radimrehurek.com/gensim/about.html)
# - Based on http://radimrehurek.com/gensim/tut1.html.
# - This handles the conversion into gensim's vector representation from raw text.
# - See perform_lsa.py for script that performs latent semantic analsis using gensim.
# - The output dictionary is compressed by default (via bzip2) as with Gensim script
#   for wikipedia formatting.
# - See google_word2vec.py for script that supports term similarity instead of document similarity.
#
# TODO:
# - Use token quoting consistent with Gensim topic display (i.e., double).
# - Work around quirk requiring the model to be saved prior to similarity calculations.
# 

"""Interface into Gensim package for vector-based text analysis (e.g., document similarity)"""

# Standard packages
import argparse
import logging
from collections import OrderedDict
import os
import re
import sys
import tempfile

# Installed packages
from gensim import corpora, models, similarities

# Local packages
from mezcla import system
import mezcla.tpo_common as tpo
from mezcla import debug
import mezcla.glue_helpers as gh

# Environment options
# Note: These are for internal options not intended for end users (e.g., for default values).
# This also allows for defining options in one place rather than say 3+ for argparse.
# TODO: XYZ => DEFAULT_XYZ
MAX_SIMILAR = tpo.getenv_integer("MAX_SIMILAR", 10)
DOCID_FILENAME = tpo.getenv_value("DOCID_FILENAME", None)
# TODO: default to number of CPU's
PARALLEL_SHARDS = tpo.getenv_integer("PARALLEL_SHARDS", 1)
IN_MEMORY = tpo.getenv_bool("IN_MEMORY", False)
TEMP_BASE = tpo.getenv_text("TEMP_BASE", tempfile.NamedTemporaryFile().name)
#
# The following are for pruning dictionary
MIN_NUM_DOCS = tpo.getenv_integer("MIN_NUM_DOCS", None)
MAX_PCT_DOCS = tpo.getenv_number("MAX_PCT_DOCS", None)
MAX_NUM_TOKENS = tpo.getenv_integer("MAX_NUM_TOKENS", None)

#------------------------------------------------------------------------

class CorpusData(object):
    """Class for processing corpora with gensim (based on MyCorpus from gensim samples)"""
    # Note: Corpora can be too large for main memory, so designed around iterators.
    # TODO: Add option for using all available memory.
    # TODO: isolate class into separate module

    def __init__(self, text=None, directory=None, in_memory=None):
        """Constructor: initialize dictionary mapping for terms"""
        tpo.debug_print("CorpusData.__init__(%s)" % text, 6)
        # TODO: self.text => self.filename
        debug.assertion(not (text and directory))
        self.text = text        # file representing entire corpus: one line per document
        self.directory = directory      # directory with corpus files
        self.mm = None          # matrix market format (see http://math.nist.gov/MatrixMarket/formats.html)
        if in_memory is None:
            in_memory = IN_MEMORY
        self.in_memory = in_memory      # keep matrix in memory
        ## OLD:
        ## if (self.text):              # mapping from words to token IDs
        ##     self.dictionary = create_dictionary(self.text)
        self.dictionary = None
        if (self.text or self.directory):
            self.create_gensim_dictionary()
        return

    def load(self, basename):
        """Load corpus model and dictionary from disk using BASENAME"""
        tpo.debug_print("CorpusData.load(%s)" % basename, 6)
        ## OLD: assert(self.text is None)
        debug.assertion((self.text is None) and (self.directory is None))
        self.dictionary = corpora.Dictionary.load_from_text(basename + '.wordids.txt.bz2')
        # TODO: make sure the document index gets loaded (for random access)
        self.mm = corpora.MmCorpus(basename + '.bow.mm')
        tpo.trace_object(self.mm, 7, "mm")
        return

    ## TODO: convert save support from client code below
    ## def save(self):
    ##   ...
    
    def read_corpus_from_file(self, filename):
        """Interates through the corpus a document at a time based on lines in FILENAME (i.e., combined data file)"""
        debug.trace_fmt(5, "read_corpus_from_file([file={f}])", f=filename)
        # Notes: old-style input; this uses previously specified text unless overridden because 
        # the main representation for corpus is the matrix (i.e., self.mm).

        # Use previous contents if not specified
        if filename is not None:
            self.text = filename
            self.directory = None

        # Read each logical file returning entire lowercased contents
        for line in system.open_file(self.text):
            # note: assumes there's one document per line with tokens separated by whitespace
            file_tokenized = re.split(r"\W+", line.lower())
            debug.trace_fmt(6, "yielding tokens: len={l}", l=len(file_tokenized))
            ## BAD: yield self.dictionary.doc2bow(file_tokenized, allow_update=update_dict)
            yield file_tokenized
        return

    def read_corpus_from_directory(self, filename):
        """Iterates through the corpus a document at a time using files in directory FILENAME. 
        Note: the files are added in lexicographical order"""
        debug.trace_fmt(5, "read_corpus_from_directory([dir={d}])", d=filename)
        # Notes: As with read_corpus_from_file, the directory name is reused in later calls.
        # Each individual file is assumed to fit in memory.

        # Use previous directory if not specified
        if filename is not None:
            self.directory = filename
            self.text = None
            
        # Read each physical file returning entire lowercased contents
        # TODO: use new system.get_directory_filenames()
        for dir_filename in sorted(system.read_directory(self.directory)):
            full_path = gh.form_path(self.directory, dir_filename)
            if not system.is_directory(full_path):
                file_tokenized = re.split(r"\W+", system.read_entire_file(full_path).lower())
                debug.trace_fmt(6, "yielding tokens: len={l}", l=len(file_tokenized))
                ## BAD: yield self.dictionary.doc2bow(file_tokenized, allow_update=update_dict)
                yield file_tokenized
        return

    def read_corpus_files(self):
        """Read corpus either from single file or directory"""
        debug.trace(5, "read_corpus_files()")
        file_arg = self.directory if self.directory else self.text
        corpus_input_fn = self.read_corpus_from_directory if self.directory else self.read_corpus_from_file
        for i, file_contents in enumerate(corpus_input_fn(file_arg)):
            ## BAD: debug.trace_fmt(6, "file contents {n}: {contents}", n=(i + 1), contents=gh.elide(file_contents))
            debug.trace_fmt(6, "yielding baw of words for file {n}: len={l}", n=(i + 1), l=len(file_contents))
            yield file_contents
        return

    def create_gensim_dictionary(self):
        """Create dictionary with word mappings and frequencies from input source (see read_corpus_files)"""
        # Note: The vectors are only retained if in-memory usage.
        # This is just done for the side effect (see gensim's dictionary.py).
        debug.trace(5, "create_gensim_dictionary()")
        self.dictionary = corpora.Dictionary()
        if self.in_memory:
            self.mm = []
        for file_contents in self.read_corpus_files():
            vector = self.dictionary.doc2bow(file_contents, allow_update=True)
            if self.in_memory:
                self.mm.append(vector)
        return
    
    def __iter__(self):
        """Returns iterator over vectors in corpus"""
        ## OLD: """Returns iterator over vectors in corpus or over lines in input text"""
        tpo.debug_print("CorpusData.__iter__()", 6)
        if (self.mm):
            for vector in self.mm.__iter__():
                yield vector
        else:
            debug.trace(4, "Warning: re-reading corpus for iter--use load() so that mm defined.")
            for file_contents in self.read_corpus_files():
                yield self.dictionary.doc2bow(file_contents, allow_update=False)
        return

    def __len__(self):
        """Returns number of documents in corpus"""
        ## OLD: num_docs = len(self.mm) if self.mm else self.text_length()
        num_docs = -1
        if (self.mm):
            num_docs = len(self.mm)
        else:
            debug.trace(4, "Warning: re-reading corpus for len--use load() so that mm defined.")
            num_docs = self.text_length()
        tpo.debug_print("CorpusData.__len__() => %d" % num_docs, 7)
        return (num_docs)

    def text_length(self):
        """Returns number of documents in corpus (n.b., re-reads the corpus data)
        Note: Result is number of lines in self.text or number of regular files in self.directory."""
        ## OLD: """Returns number of documents (i.e., lines) in text"""
        length = 0
        ## OLD: for _line in open(self.text):
        for _line in self.read_corpus_files():
            length += 1
        tpo.debug_print("CorpusData.text_length() => %d" % length, 7)
        return (length)

    def __getitem__(self, index):
        """Returns corpus item at INDEX (0-based)"""
        result = None
        try:
            result = self.mm[index]
        except RuntimeError:
            tpo.print_stderr("Warning: falling back to linear document access (i.e., not random access): make sure index file exists")
            result = None
            for i, value in enumerate(self.mm):
                if (i == index):
                    result = value
                    break
        return (result)

#------------------------------------------------------------------------

class UserIdMapping(object):
    """Class for mapping Gensim document ID's into user ID's"""

    def __init__(self, docid_filename):
        """Class constructor"""
        ## OLD: self.docid_mapping = tpo.create_lookup_table(docid_filename, use_linenum=True)
        self.docid_mapping = create_ordered_lookup_table(docid_filename)
        # TODO: just use array indexing (e.g., doc_positions = docid_mapping.values())
        ## OLD: self.reverse_docid_mapping = dict((docid, key) for (key, docid) in list(self.docid_mapping.items()))
        self.reverse_docid_mapping = dict((label, docid) for (docid, label) in list(self.docid_mapping.items()))
        return

    def get_user_id(self, docid):
        """Returns user ID for DOCID (same as input if no mapping exists)"""
        # Note: user_id is user's ID for document (not gensim's)
        if (not isinstance(docid, str)):
            docid = str(docid)
        user_id = self.docid_mapping.get(docid, docid)
        gh.assertion(user_id != docid)
        ## OLD: tpo.debug_format("get_user_id({docid}) => {user_id}", 6)
        debug.trace_fmt(6, "get_user_id({d}) => {u}", d=docid, u=user_id)
        return user_id

    def get_gensim_id(self, docid):
        """Returns gensim ID for DOCID (same as input if no mapping exists)"""
        if (not isinstance(docid, str)):
            docid = str(docid)
        gensim_id = self.reverse_docid_mapping.get(docid, docid)
        gh.assertion(gensim_id != docid)
        ## OLD: tpo.debug_format("get_gensim_id({docid}) => {gensim_id}", 6)
        debug.trace_fmt(6, "get_gensim_id({d}) => {g}", d=docid, g=gensim_id)
        return gensim_id

#------------------------------------------------------------------------

class SimilarDocument(object):
    """Base class for finding similar documents via vector-space cosine measure"""
    # TODO: rework so that CorpusData class used to encapsulate both corpus and dictionary

    def __init__(self, corpus=None, dictionary=None, verbose_output=False, max_similar=MAX_SIMILAR, docid_filename=DOCID_FILENAME):
        """Class constructor"""
        ## OLD: tpo.debug_format("SimilarDocument.__init__({corpus}, {dictionary}, {verbose_output}, {max_similar}, {docid_filename})", 6)
        debug.trace_fmt(6, "SimilarDocument.__init__({c}, {d}, {v}, {ms}, {df})", c=corpus, d=dictionary, v=verbose_output, ms=max_similar, df=docid_filename)
        # If specified, override the number of CPUs for parallel processing of shards
        if (PARALLEL_SHARDS > 1):
            similarities.docsim.PARALLEL_SHARDS = PARALLEL_SHARDS
        self.corpus = corpus
        self.dictionary = dictionary
        self.sim_index = None
        if self.corpus:
            ## OLD: gh.assertion(isinstance(self.corpus, corpora.MmCorpus))
            gh.assertion(isinstance(self.corpus, corpora.MmCorpus) or all_lists_of_num_pairs(self.corpus))
        if self.dictionary:
            gh.assertion(isinstance(self.dictionary, corpora.Dictionary))
        self.verbose_output = verbose_output
        self.max_similar = max_similar
        self.docid_mapping = UserIdMapping(docid_filename) if docid_filename else None
        return

    def get_user_id(self, docid):
        """Returns user ID for DOCID (same as input if no mapping exists)"""
        return (self.docid_mapping.get_user_id(docid) if self.docid_mapping else docid)

    def get_gensim_id(self, docid):
        """Returns gensim ID for DOCID (same as input if no mapping exists)"""
        return (self.docid_mapping.get_gensim_id(docid) if self.docid_mapping else docid)

    def find(self, _docid):
        """"Return documents similar to DOCID; result is a list of tuples: (docid, weight)"""
        tpo.debug_format("SimilarDocument.find({d}); self={s}", 6,
                         d=_docid, s=self)
        assert(False)
        return []

    def find_all_similar(self):
        """Iterator for getting list of similar documents for each document: each result is a tuple (docid, similar-doc-list), with similar-doc-list a list of (other-docid, weight) tuples"""
        tpo.debug_print("SimilarDocument.find_all_similar", 5)
        for docid in range(len(self.corpus)):
            yield(docid, self.find(docid))
        return

    def load(self, filename):
        """Loads similarity model from FILENAME"""
        # Note: for interactive usage only
        tpo.debug_print("Loading similiarity index from FILENAME", 4)
        self.sim_index = similarities.Similarity.load(filename)
        return

    def save(self, filename):
        """Saves similarity model to FILENAME"""
        ## OLD: tpo.debug_format("Saving similiarity index to {filename}", 4)
        debug.trace_fmt(4, "Saving similiarity index to {f}", f=filename)
        return (self.sim_index.save(filename))


class SimilarDocumentByCosine(SimilarDocument):
    """Class for finding similar documents via vector-space cosine measure"""

    def __init__(self, corpus=None, dictionary=None, index_file=None, verbose_output=False, max_similar=MAX_SIMILAR, docid_filename=DOCID_FILENAME):
        """Class constructor"""
        ## OLD: tpo.debug_format("SimilarDocumentByCosine.__init__({corpus}, {dictionary}, {index_file}, {verbose_output}, {max_similar}, {docid_filename})", 6)
        debug.trace_fmt(6, "SimilarDocumentByCosine.__init__(corp={c}, dict={d}, indf={f}, verb={v}, maxsim={ms}, docfil={df})", c=corpus, d=dictionary, f=index_file, v=verbose_output, ms=max_similar, df=docid_filename)
        # note: index_file serves both as the cache for the similarity object as well as base name for the shards it uses (see gensim's docsim.py)
        # TODO: rework so that corpus and dictionary not needed to retrieve pre-computed similarity results
        SimilarDocument.__init__(self, corpus, dictionary, verbose_output, max_similar, docid_filename)
        ## BAD: if (self.corpus and self.dictionary):
        if ((self.corpus is not None) and (self.dictionary is not None)):
            if index_file is None:
                index_file = TEMP_BASE + "-simindex"
            if (gh.non_empty_file(index_file)):
                self.sim_index = similarities.Similarity.load(index_file)
                # Make sure shard file prefix matches index file
                if self.sim_index.output_prefix != index_file:
                    tpo.debug_format("Updating shard file prefix: {self.sim_index.output_prefix} => {index_file}", 5)
                    self.sim_index.output_prefix = index_file
                    self.sim_index.check_moved()
            else:
                self.sim_index = similarities.Similarity(index_file, self.corpus, len(self.dictionary), max_similar)
            tpo.debug_print("sim_index: type=%s value=%s" % (type(self.sim_index), self.sim_index), 5)
        else:
            ## OLD: tpo.debug_format("c={self.corpus} d={self.dictionary} sim_index={self.sim_index}", 6)
            debug.trace_fmt(5, "c={c} d={d} sim_index={i}", c=self.corpus, d=self.dictionary, i=self.sim_index)
        tpo.trace_object(self, 5, "SimilarDocumentByCosine.self")
        return

    def normalize_score(self, score):
        """Normalized from cosine value in range [-1, 1] to probability-type score in range [0, 1])"""
        # pylint: disable=no-self-use
        MIN_SCORE = -1.0
        MAX_SCORE = 1.0
        EPSILON = 0.001
        gh.assertion(MIN_SCORE-EPSILON <= score <= MAX_SCORE+EPSILON)
        debug.assertion(MIN_SCORE < MAX_SCORE)
        normal_score = (float(score - MIN_SCORE) / (MAX_SCORE - MIN_SCORE))
        ## tpo.debug_format("SimilarDocumentByCosine.normalize_score({score}) => {normal_score}", 6)
        debug.trace_fmt(6, "SimilarDocumentByCosine.normalize_score({s}) => {ns}",
                        s=score, ns=normal_score)
        return (normal_score)

    def find(self, docid):
        """"Return documents similar to DOCID; result is a list of tuples: (docid, weight)"""
        ## tpo.debug_format("SimilarDocumentByCosine.find({docid})", 5)
        debug.trace_fmt(5, "SimilarDocumentByCosine.find({d})", d=docid)
        debug.trace_fmt(5, "corpus/dict/index={c}/{d}/{i}", c=self.corpus, d=self.dictionary, i=self.sim_index)
        gh.assertion(self.corpus and self.dictionary and self.sim_index)
        ## OLD: gensim_docid = self.get_gensim_id(docid)
        gensim_docid = docid
        try:
            similar_gensim_docs = self.sim_index[self.corpus[int(gensim_docid)]]
            similar_docs = [(self.get_user_id(doc), self.normalize_score(score)) for (doc, score) in similar_gensim_docs]
            if self.verbose_output:
                similar_docs = [(docid, score, resolve_terms(docid, self.dictionary)) for (docid, score) in similar_docs]
        except:
            tpo.debug_raise()
            tpo.print_stderr("Exception retrieving similar documents: " + str(sys.exc_info()))
            similar_docs = []
        result = similar_docs
        ## OLD: tpo.debug_format("find({docid}) => {result}", 5)
        debug.trace_fmt(5, "find({d}) => {r}", d=docid, r=result)
        return result

    def derive_all_similarities(self):
        """Precompute similarities, using batch method via chunking (see Gensim documentation in docsim.py)."""
        tpo.debug_format("SimilarDocumentByCosine.derive_all_similarities()", 5)
        _all_sim = list(self.sim_index[self.corpus])
        return

#------------------------------------------------------------------------

def create_dictionary(filename):
    """Create dictionary with mord mappings and frequencies from FILENAME with each line representing a separate document"""
    # Note: Updates the dictionary for each line (to allow for very large corpora files).
    tpo.debug_print("create_dictionary(%s)" % filename, 5)
    dictionary = corpora.Dictionary()
    ## OLD: for line in open(filename):
    for line in system.open_file(filename):
        ## OLD: line_tokenized = [w for w in re.split(r"\W+", line)]
        line_tokenized = re.split(r"\W+", line)
        dictionary.doc2bow(line_tokenized, allow_update=True)
    return (dictionary)


def create_ordered_lookup_table(filename):
    """Create lookup hash table from 0-based line number as string to labels specified in FILENAME
    Note: case is preserved but leading and trailing whitespace is removed"""
    tpo.debug_print("create_ordered_lookup_table(%s)" % filename, 5)
    lookup_hash = OrderedDict()
    f = None
    try:
        f = system.open_file(filename)
        line_num = 0
        for line in f:
            lookup_hash[str(line_num)] = line.strip()
            line_num += 1
    except (IOError, ValueError):
        tpo.debug_print("Warning: Exception creating lookup table from %s: %s" % (filename, str(sys.exc_info())), 2)
    finally:
        if f:
            f.close()
    tpo.debug_print("create_ordered_lookup_table() => %s" % lookup_hash, 6)
    return lookup_hash


def resolve_terms(vector, dictionary):
    """Return vector with token ID's replaced by the actual tokens"""
    tpo.debug_print("resolve_terms(%s, %s)" % (vector, dictionary), 7)
    term_vector = [(dictionary[token_id], count) for (token_id, count) in vector]
    ## OLD: return sorted(term_vector, reverse=True, key=lambda _term, freq: freq)
    return sorted(term_vector, reverse=True, key=lambda t_f: t_f[1])


def all_num_pairs(l):
    """Whether L is list of tuples (n1, n2)"""
    result = (isinstance(l, list)
              and all(map(lambda t: (len(t) == 2) and tpo.is_numeric(t[0]) and tpo.is_numeric(t[1]),
                          l)))
    debug.trace_fmt(7, "all_num_pairs({v}) => {r}", v=gh.elide(str(l)), r=result)
    return result


def all_lists_of_num_pairs(l):
    """Whether each item in L is a list of numeric tuples"""
    result = isinstance(l, list) and all(map(all_num_pairs, l))
    debug.trace_fmt(6, "all_lists_of_num_pairs({v}) => {r}", v=gh.elide(str(l)), r=result)
    return result

#-------------------------------------------------------------------------------

def main(runtime_args=None):
    """Entry point for script"""
    # Note: Uses sys.argv if no arguments supplied
    ## OLD: tpo.debug_print("main(): args=%s" % sys.argv, 4)
    debug.trace_fmt(5, "main(): args=[{ra}]", ra=runtime_args)
    ## TODO
    ## if (runtime_args is None):
    ##     runtime_args = sys.argv
    ##     debug.trace_fmt(4, "Using sys.argv for args: {ra}", ra=runtime_args)
    if (runtime_args is not None):
        debug.trace_fmt(2, "Warning: Setting sys.argv from args: {ra}", ra=runtime_args)

    # Check command-line arguments
    # TODO: make sure each argument supported via external API
    parser = argparse.ArgumentParser(description="Creates gensim corpus files optionally with TF/IDF weighting (for use with perform_lsa.py. Note: input should be a text file when creating from scratch or the basename of model files if loading existing model(s).")
    #
    parser.add_argument("--save", default=False, action='store_true', help="Save model(s) to disk")
    parser.add_argument("--load", default=False, action='store_true', help="Load model(s) from disk")
    parser.add_argument("--tfidf", default=False, action='store_true', help="Include TF/IDF analysis")
    parser.add_argument("--original", default=False, action='store_true', help="Output original document term matrix (i.e., non-tfidf) when --tfidf specified")
    parser.add_argument("--similarity", default=False, action='store_true', help="Derive similarity data")
    parser.add_argument("--print", default=False, action='store_true', help="Print vectors on standard output")
    parser.add_argument("--expand", default=False, action='store_true', help="Expand corpus in memory")
    parser.add_argument("--verbose", default=False, action='store_true', help="Verbose output mode (e.g., resolve term ID's)")
    parser.add_argument("--normalize", default=True, action='store_true', help="Normalize TF/IDF scores")
    parser.add_argument("--skip-normalize", dest='normalize', action='store_false', help="Normalize TF/IDF scores")
    parser.add_argument("--similar-docs-of", default="", help="Show similar documents for list of document ID's (or * for all); note: currently requires --load")
    parser.add_argument("--max-similar", type=int, default=MAX_SIMILAR, help="Maximum number of similar documents to return")
    parser.add_argument("--output-basename", default="", help="Basename to use for output (by default input file without .txt extension)")
    parser.add_argument("--docid-filename", default=None, help="Filename with document ID's")
    parser.add_argument("--prune-dictionary", default=False, action='store_true', help="Prune dictionary of low/high frequency terms")
    #
    # note: filename is positional argument
    parser.add_argument("filename", default=None, help="Input data filename (or basename when loading previously saved model); use - for stdin")
    #
    ## OLD: args = vars(parser.parse_args())
    args = vars(parser.parse_args(runtime_args))
    tpo.debug_print("args = %s" % args, 5)
    filename = args['filename']
    save = args['save']
    load = args['load']
    perform_tfidf = args['tfidf']
    verbose_output = args['verbose']
    normalize = args['normalize']
    print_vectors = args['print']
    show_original = (args['original'] or (print_vectors and (not perform_tfidf)))
    expand_corpus = args['expand']
    derive_similarity = args['similarity']
    max_similar = args['max_similar']
    source_similar_docs = args['similar_docs_of'].replace(",", " ").split()
    output_basename = args['output_basename']
    docid_filename = args['docid_filename']
    prune_dictionary = args['prune_dictionary']
    temp_file = None

    # Map stdin to temporary file
    # TODO: rework in terms of streaming (e.g., for files > 4gb)
    if (filename == "-"):
        tpo.debug_print("Reading input from stdin", 4)
        ## OLD:
        ## temp_file = tpo.getenv_text("TEMP_FILE", tempfile.NamedTemporaryFile().name)
        temp_file = TEMP_BASE
        filename = temp_file + ".txt"
        gh.write_file(filename, gh.read_file(None))

    # Derive the basename if not given
    # TODO: TFIDF_MODEL_EXT = ".tfidf.mm"
    matrix_ext = ".tfidf.mm" if perform_tfidf else ".bow.mm"
    input_extension = ".txt" if (not load) else matrix_ext
    if not output_basename:
        output_basename = gh.remove_extension(filename, input_extension)
    tpo.debug_print("output_basename=%s" % output_basename, 5)

    # Make sure input file exists
    if (not os.path.exists(filename)) and (not load):
        if (os.path.exists(filename + input_extension)):
            filename += input_extension

    # Enable logging if debugging
    # TODO: tpo.init_logging()
    if (tpo.debugging_level()):
        # TODO: use mapping from symbolic LEVEL user option (e.g., via getenv)
        level = logging.INFO if (tpo.debug_level < 4) else logging.DEBUG
        logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=level)

    # Read in corpus, optionally in saved vector format
    # TODO: put load/save support into CorpusData (and isolate as gensim_corpus.py)
    # TODO: bypass loading if TF/IDF version of corpus is being used for document similarity
    if load:
        corpus_basename = tpo.getenv_text("CORPUS_BASENAME", filename)
        corpus_data = CorpusData()
        corpus_data.load(corpus_basename)
    else:
        ## OLD: corpus_data = CorpusData(filename)
        is_dir = system.is_directory(filename)
        corpus_data = CorpusData(directory=filename) if is_dir else CorpusData(text=filename)
    tpo.debug_print("corpus_data: type=%s value=%s" % (type(corpus_data), corpus_data), 5)
    debug.trace_object(5, corpus_data)

    # Optionally prune low and high frequency terms from dictionary
    if (prune_dictionary or MIN_NUM_DOCS or MAX_PCT_DOCS or MAX_NUM_TOKENS):
        tpo.debug_print("Pruning dictionary", 4)
        option_overrides = {}
        if MIN_NUM_DOCS:
            option_overrides['no_below'] = MIN_NUM_DOCS
        if MAX_PCT_DOCS:
            option_overrides['no_above'] = MAX_PCT_DOCS
        if MAX_NUM_TOKENS:
            option_overrides['keep_n'] = MAX_NUM_TOKENS
        corpus_data.dictionary.filter_extremes(**option_overrides)

    # Print the corpus
    if (print_vectors and show_original):
        print("corpus_data: [")
        for docid, vector in enumerate(corpus_data):
            print(docid, vector if not verbose_output else resolve_terms(vector, corpus_data.dictionary))
        print("]")

    # Do optional TF/IDF analysis
    # Note: tfidf represents the transformation and tfidf_corpus the transformed corpus
    # See http://radimrehurek.com/gensim/tut2.html#transformation-interface.
    # TODO: put TFIDF support into CorpusData class
    if (perform_tfidf):
        tpo.debug_print("TF/IDF", 6)
        if load:
            tfidf_corpus = corpora.MmCorpus(output_basename + '.tfidf.mm')
        else:
            mm = list(corpus_data) if expand_corpus else corpus_data
            ## TODO: mm = corpora.MmCorpus(output_basename + '.bow.mm')
            tfidf = models.TfidfModel(mm, id2word=corpus_data.dictionary, normalize=normalize)
            tfidf_corpus = tfidf[corpus_data]
        ## TODO: tfidf = models.TfidfModel(corpus_data)
        tpo.debug_print("tfidf: type=%s value=%s" % (type(tfidf_corpus), tfidf_corpus), 5)

        # Print the TF/IDF version of the corpus
        if (print_vectors):
            print("tfidf corpus: [")
            for docid, vector in enumerate(tfidf_corpus):
                print(docid, vector if not verbose_output else resolve_terms(vector, corpus_data.dictionary))
            print("]")

    # Determine similarity model
    if (derive_similarity or source_similar_docs):
        assert load, "similarity support requires --load (to use Matrix Market format produced by --save)"
        sim_corpus = tfidf_corpus if perform_tfidf else corpus_data.mm
        dictionary = corpus_data.dictionary
        index_filename = tpo.getenv_text("SIM_INDEX", output_basename + ".sim_index")
        sim = SimilarDocumentByCosine(corpus=sim_corpus, dictionary=dictionary, index_file=index_filename, verbose_output=verbose_output, max_similar=max_similar, docid_filename=docid_filename)
        # Precompute similarities
        if not source_similar_docs:
            sim.derive_all_similarities()

    # Show similar documents
    # TODO: have option to save as data file
    if source_similar_docs:
        if (source_similar_docs == ['*']):
            similar_doc_info = sim.find_all_similar()
        else:
            similar_doc_info = []
            for docid in source_similar_docs:
                similar_doc_info.append((docid, sim.find(docid)))
                ## TODO: 
                ## user_docid = sim.get_user_id(docid)
                ## similar_doc_info.append((user_docid, sim.find(docid)))
        if tpo.verbose_debugging():
            similar_doc_info = list(similar_doc_info)
            ## OLD: tpo.debug_format("similar_doc_info={similar_doc_info}")
            debug.trace_fmt(4, "similar_doc_info={sdi}", sdi=similar_doc_info)
        for (docid, similar_docs) in similar_doc_info:
            similar_docs = [(d, tpo.round_num(score)) for (d, score) in similar_docs]
            ## OLD: print("Documents similar to %s: %s" % (docid, similar_docs))
            ## TODO: remove source docid's (i.e., no identity matches)
            user_docid = sim.get_user_id(docid)
            other_similar_docs = [(d, s) for (d, s) in similar_docs if (d != user_docid)]
            print("Documents similar to %s: %s" % (user_docid, other_similar_docs))
            ## TODO: show overlapping terms in verbose mode

    # Optionally save main components to disk
    if (save):
        # TODO: have corpus data save both the dictionary and the matrix
        tpo.debug_print("saving corpora files", 6)
        if ((not load) or (not gh.non_empty_file(output_basename + '.wordids.txt.bz2'))):
            corpus_data.dictionary.save_as_text(output_basename + '.wordids.txt.bz2')
        if ((not load) or (not gh.non_empty_file(output_basename + '.bow.mm'))):
            mm = list(corpus_data) if expand_corpus else corpus_data
            corpora.MmCorpus.serialize(output_basename + '.bow.mm', mm)
        if (perform_tfidf):
            if ((not load) or (not gh.non_empty_file(output_basename + '.tfidf.mm'))):
                corpora.MmCorpus.serialize(output_basename + '.tfidf.mm', tfidf_corpus)
        if (derive_similarity):
            sim.save(output_basename + '.sim_index')

    # Cleanup
    if (temp_file and (not tpo.detailed_debugging())):
        ## OLD: gh.run("rm -vf {temp_file}")
        gh.run("rm -vf {temp_file}*")

    return

#------------------------------------------------------------------------

if __name__ == '__main__':
    ## OLD:
    main()
    ## TODO: main(sys.argv)

## TEST (for testing tests/test_gensim_test.py conditional loading):
## fubar



#! /usr/bin/env python
#
# Tests for transpose_data module
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_transpose_data.py
#

"""Tests for transpose_data module"""

# Standard modules
## NOTE: this is empty for now

# Installed modules
import pytest

# Local modules
from mezcla import debug

# Note: Rreference are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.transpose_data as THE_MODULE

class TestTransposeData:
    """Class for testcase definition"""

    ## TODO: TESTS WORK-IN-PROGRESS


if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# Test(s) for ../text_processing.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_text_processing.py
#

"""Tests for text_processing module"""

# Standard packages
## NOTE: this is empty for now

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.text_processing as THE_MODULE
from mezcla.unittest_wrapper import TestWrapper

# Constants
RESOURCES = f'{gh.dir_path(__file__)}/resources'
TEXT_EXAMPLE = f'{RESOURCES}/example_text.txt'
TEXT_EXAMPLE_TAGS = f'{RESOURCES}/example_text_tags.txt'
WORD_POS_FREQ_FILE = f'{RESOURCES}/word-POS.freq'

class TestTextProcessing:
    """Class for testcase definition"""

    def test_split_sentences(self):
        """Ensure split_sentences works as expected"""
        debug.trace(4, "test_split_sentences()")
        assert THE_MODULE.split_sentences("I came. I saw. I conquered!") == ["I came.", "I saw.", "I conquered!"]
        assert THE_MODULE.split_sentences("Dr. Watson, it's elementary. But why?") == ["Dr. Watson, it's elementary.", "But why?"]

    def test_split_word_tokens(self):
        """Ensure split_word_tokens works as expected"""
        debug.trace(4, "test_split_word_tokens()")
        assert THE_MODULE.split_word_tokens("How now, brown cow?") == ['How', 'now', ',', 'brown', 'cow', '?']

    def test_label_for_tag(self):
        """Ensure label_for_tag works as expected"""
        debug.trace(4, "test_label_for_tag()")
        ## TODO: WORK-IN=PROGRESS

    def test_class_for_tag(self):
        """Ensure class_for_tag works as expected"""
        debug.trace(4, "test_class_for_tag()")
        assert THE_MODULE.class_for_tag("NNS") == "noun"
        assert THE_MODULE.class_for_tag("VBG") == "verb"
        assert THE_MODULE.class_for_tag("VBG", previous="MD") == "verb"
        assert THE_MODULE.class_for_tag("NNP", word="(") == "punctuation"

    def test_tag_part_of_speech(self):
        """Ensure tag_part_of_speech works as expected"""
        debug.trace(4, "test_tag_part_of_speech()")
        # NOTE:
        #   'brown' tagged as IN is wrong, should be JJ, this is a problem related to NLTK
        #   not the module being tested, so we are ignoring it for now.
        #   Related: https://stackoverflow.com/a/30823202
        assert THE_MODULE.tag_part_of_speech(['How', 'now', ',', 'brown', 'cow', '?']) == [('How', 'WRB'), ('now', 'RB'), (',', ','), ('brown', 'IN'), ('cow', 'NN'), ('?', '.')]

    def test_tokenize_and_tag(self):
        """Ensure tokenize_and_tag works as expected"""
        debug.trace(4, "test_tokenize_and_tag()")
        ## TODO: WORK-IN=PROGRESS

    def test_tokenize_text(self):
        """Ensure tokenize_text works as expected"""
        debug.trace(4, "test_tokenize_text()")
        ## TODO: WORK-IN=PROGRESS

    def test_is_stopword(self):
        """Ensure is_stopword works as expected"""
        debug.trace(4, "test_is_stopword()")
        ## TODO: WORK-IN=PROGRESS

    def test_has_spelling_mistake(self):
        """Ensure has_spelling_mistake works as expected"""
        debug.trace(4, "test_has_spelling_mistake()")
        ## TODO: WORK-IN=PROGRESS

    def test_read_freq_data(self):
        """Ensure read_freq_data works as expected"""
        debug.trace(4, "test_read_freq_data()")
        ## TODO: WORK-IN=PROGRESS

    def test_read_word_POS_data(self): # pylint: disable=invalid-name
        """Ensure read_word_POS_data works as expected"""
        debug.trace(4, "test_read_word_POS_data()")
        ## TODO: WORK-IN=PROGRESS

    def test_get_most_common_POS(self): # pylint: disable=invalid-name
        """Ensure get_most_common_POS works as expected"""
        debug.trace(4, "test_get_most_common_POS()")
        # Testing if word_POS_hash is not none
        THE_MODULE.word_POS_hash = {
            'can': 'MD'
        }
        assert THE_MODULE.get_most_common_POS("notaword") == "NN"
        assert THE_MODULE.get_most_common_POS("can") == "MD"
        # Test reading from freq file
        ## TODO: for some reason this assertion freezes
        ## THE_MODULE.WORD_POS_FREQ_FILE = 'tests/resources/word-POS.freq'
        ## assert THE_MODULE.get_most_common_POS("to") == "TO"

    def test_is_noun(self):
        """Ensure is_noun works as expected"""
        debug.trace(4, "test_is_noun()")
        assert THE_MODULE.is_noun('notaword', 'NN')
        assert not THE_MODULE.is_noun('can', 'MD')

    def test_is_verb(self):
        """Ensure is_verb works as expected"""
        debug.trace(4, "test_is_verb()")
        assert THE_MODULE.is_verb('run', 'VB')
        assert not THE_MODULE.is_verb('can', 'NN')

    def test_is_adverb(self):
        """Ensure is_adverb works as expected"""
        debug.trace(4, "test_is_adverb()")
        assert THE_MODULE.is_adverb('quickly', 'RB')
        assert not THE_MODULE.is_adverb('can', 'MD')

    def test_is_adjective(self):
        """Ensure is_adjective works as expected"""
        debug.trace(4, "test_is_adjective()")
        assert THE_MODULE.is_adjective('quick', 'JJ')
        assert not THE_MODULE.is_adjective('can', 'MD')

    def test_is_comma(self):
        """Ensure is_comma works as expected"""
        debug.trace(4, "test_is_comma()")
        assert THE_MODULE.is_comma('comma', ',')
        assert THE_MODULE.is_comma(',', 'comma')
        assert not THE_MODULE.is_comma('can', 'MD')

    def test_is_quote(self):
        """Ensure is_quote works as expected"""
        debug.trace(4, "test_is_quote()")
        assert THE_MODULE.is_quote('\'', '')
        assert THE_MODULE.is_quote('\"', '')
        assert not THE_MODULE.is_quote('can', 'MD')

    def test_is_punct(self):
        """Ensure is_punct works as expected"""
        debug.trace(4, "test_is_punct()")
        assert THE_MODULE.is_punct('$', '$')
        assert not THE_MODULE.is_punct('can', 'MD')

    def test_usage(self):
        """Ensure usage works as expected"""
        debug.trace(4, "test_usage()")
        ## TODO: WORK-IN=PROGRESS


class TestTextProcessingScript(TestWrapper):
    """Class for testcase definition"""
    script_file = TestWrapper.get_module_file_path(__file__)
    script_module =TestWrapper.get_testing_module_name(__file__)

    def test_all(self):
        """Ensure text_processing without argument works as expected"""
        debug.trace(4, "test_all()")
        output = self.run_script(data_file=TEXT_EXAMPLE)
        assert output == gh.read_file(TEXT_EXAMPLE_TAGS)[:-1]

    def test_just_tokenize(self):
        """Ensure just_tokenize argument works as expected"""
        debug.trace(4, "test_just_tokenize()")
        ## TODO: WORK-IN-PROGRESS

    def test_make_lowercase(self):
        """Ensure make_lowercase argument works as expected"""
        debug.trace(4, "test_make_lowercase()")
        ## TODO: WORK-IN-PROGRESS


if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# Audio Module
#
# This provides an Audio wraper interface to extract
# multiples properties that you can implement on your
# projects and also provides an argument processing class
# to easily use from terminal to use different features.
#
#
# Python code example using Audio wraper interface:
# | some_audio = Audio(path="./path/to/audio.wav")
# | print(some_audio.speech_to_text(engine="sphinx"))
#
# Example using the terminal:
# $ python audio.py path/to/folder
# $ python audio.py --speech sphinx path/to/audio.wav
#
#
# The Audio wraper provides three engines to extract speech:
# - CMUSphinx  (aka "sphinx")
# - Houndify   (aka "houndify")
# - IBM Watson (aka "watson")
#
# You can find the UML diagram related to this module on:
# - ./docs/audio_uml.svg
#


"""
Audio Module

This provides an Audio wraper interface to extract
multiples properties that you can implement on your
projects and also provides an argument processing class
to easily use from terminal to use different features.

Python code example using Audio wraper interface:
| some_audio = Audio(path="./path/to/audio.wav")
| print(some_audio.speech_to_text(engine="sphinx"))

Example using the terminal:
$ python audio.py path/to/folder
$ python audio.py --speech sphinx path/to/audio.wav

The Audio wraper provides for now three engines to extract speech:
- CMUSphinx  (aka "sphinx")
- Houndify   (aka "houndify")
- IBM Watson (aka "watson")
"""


# Standard packages
import re


# Installed packages
import librosa
import speech_recognition as sr
from ibm_watson import SpeechToTextV1
from ibm_cloud_sdk_core.authenticators import IAMAuthenticator


# Local packages
from mezcla.main import Main
from mezcla      import debug
from mezcla      import glue_helpers as gh


# Command-line labels constants
SOURCE  = 'source'
OUTPUT  = 'output'
SPEECH  = 'speech'
VERBOSE = 'verbose'


# Constants
AUDIO_FORMATS = ['aac', 'au', 'flac', 'm4a', 'mp3', 'ogg', 'wav']


# Authentications.
# This is optional; empty values must be completed with
# your credentials if you want to use some of these platforms.
IBM_WATSON_KEY = ''
IBM_WATSON_URL = ''
HOUNDIFY_ID  = ''
HOUNDIFY_KEY = ''


class AudioArgumentProcessing(Main):
    """Argument processing class"""


    # Class-level member variables for arguments (avoids need for class constructor)
    audios  = []
    output  = ''
    speech  = ''
    verbose = False


    def setup(self):
        """Process arguments"""
        debug.trace(7, f'AudioArgumentProcessing.setup({self})')

        # Check the command-line options
        source       = self.get_parsed_argument(SOURCE, "")
        self.output  = self.get_parsed_option(OUTPUT, self.output)
        self.speech  = self.get_parsed_option(SPEECH, self.speech)
        self.verbose = self.get_parsed_option(VERBOSE)

        # Process source
        source_match = re.search(r'(?P<path>[\w\-\/]+)\.*(?P<ext>[a-zA-Z0-9]+)* *$', source)
        extension = source_match.group('ext') if source_match else ''
        path      = source_match.group('path') if source_match else ''

        # Check if is a valid audio file
        if extension in AUDIO_FORMATS:
            self.audios = [source]

        # Check if is a list of audios
        elif extension == 'list':
            self.audios = gh.read_lines(source)

        # Check if is a folder
        elif path and not extension:
            self.audios = librosa.util.find_files(source, ext=AUDIO_FORMATS, recurse=True)

        else:
            print('Not valid audio files founded.')

        # Initializate audios
        for index, audio_path in enumerate(self.audios):
            self.audios[index] = Audio(path = audio_path)

        debug.trace(7, f'AudioArgumentProcessing.setup() - audios founded from ({source}): {self.audios}')


    def run_main_step(self):
        """Process main script"""
        debug.trace(7, f'AudioArgumentProcessing.run_main_step({self})')


        for current_audio in self.audios:
            debug.trace(7, f'AudioArgumentProcessing.run_main_step() - processing {current_audio}')

            output = ''

            if self.verbose:
                output += f'Audio: {current_audio.get_path()}\n'

            # Extract speech
            if self.speech:
                if self.verbose:
                    output += '- speech recognized:\n'
                output += current_audio.speech_to_text(engine = self.speech)
                output += '\n'

            print(output.rstrip('\n'))

            # Save on output
            ## TODO: WORK-IN-PROGRESS


class SpeechEngine:
    """SpeechEngine abstract class"""


    # Global state
    _IDENTIFIER = 'generic'


    def get_identifier(self):
        """Returns engine identifier string"""
        debug.trace(7, f'SpeechEngine.get_identifier() => {self._IDENTIFIER}')
        return self._IDENTIFIER


    def check_credentials(self):
        """Check if credentials are valid"""

        is_valid = True
        ## TODO: WORK-IN-PROGRESS

        debug.trace(7, f'SpeechEngine.check_credentials() => {is_valid}')
        return is_valid


    # Abstract method
    # pylint: disable=no-self-use
    def particular_check_audio(self, path):
        """Check if audio in PATH is valid"""
        is_valid = True
        ## OVERRIDE THIS
        debug.trace(7, f'SpeechEngine.particular_check_audio({path}) => {is_valid}')
        return is_valid


    # Abstract method
    # pylint: disable=no-self-use
    def particular_speech_to_text(self, path):
        """Extract speech from audio in PATH using particular speech engine"""
        result = ''
        ## OVERRIDE THIS
        debug.trace(7, f'SpeechEngine.particular_speech_to_text({path}) => {result}')
        return result


    def speech_to_text(self, path):
        """Transcribe speech from audio in PATH"""

        result = ''

        if not self.check_credentials():
            print(f'Invalid credentials for {self._IDENTIFIER}.')
            return result

        if not self.particular_check_audio(path):
            print(f'Invalid audio ({path}) for {self._IDENTIFIER}.')
            return result

        result = self.particular_speech_to_text(path)

        debug.trace(7, f'SpeechEngine.speech_to_text({path}) => {result}')
        return result


class IBMWatson(SpeechEngine):
    """
    Speech recognition engine for IBM Watson.
    Docs: https://cloud.ibm.com/apidocs/speech-to-text?code=python#recognize
    """
    ## TODO: solve getting very short responses (not full transcription)


    # Global state
    _IDENTIFIER = 'watson'


    def __init__(self, api_key='', api_url=''):
        self.api_key = api_key
        self.api_url = api_url


    def particular_speech_to_text(self, path):
        """Transcribe speech from audio in PATH"""
        result = ''

        # Get extension
        extension = ''
        for ext in AUDIO_FORMATS:
            if f'.{ext}' in path:
                extension = ext

        # Setup Service
        authenticator = IAMAuthenticator(self.api_key)
        stt = SpeechToTextV1(authenticator=authenticator)
        stt.set_service_url(self.api_url)

        # Perform conversion
        with open(path, 'rb') as path_audio:
            res = stt.recognize(audio=path_audio,
                                content_type=f'audio/{extension}',
                                model='en-US_NarrowbandModel',
                                inactivity_timeout=30).get_result()

        ## result = res['results'][0]['alternatives'][0]['transcript']
        result = str(res)

        debug.trace(7, f'SpeechEngine.speech_to_text({path}) => {result}')
        return result


class SpeechRecognition(SpeechEngine):
    """SpeechRecognition abstract class for speech_recognition module"""


    # Global state
    _recognition = sr.Recognizer()


    def get_instance(self):
        """Returns recognition instance"""
        return self._recognition


    def read_audio(self, path):
        """Read audio file"""

        with sr.AudioFile(path) as source:
            audio = self.get_instance().record(source)

        debug.trace(7, 'SpeechEngine.read_audio()')
        return audio


class CMUSphinx(SpeechRecognition):
    """CMUSphinx Speech Recognition class"""


    # Global state
    _IDENTIFIER = 'sphinx'


    def particular_check_audio(self, path):
        """Check if audio in PATH is valid"""
        is_valid = True

        ## TODO: WORK-IN-PROGRESS
        ## NOTE: sphinx not sopport mp3 format

        debug.trace(7, f'SpeechEngine.particular_check_audio({path}) => {is_valid}')
        return is_valid


    def particular_speech_to_text(self, path):
        """Transcribe speech from audio in PATH"""

        result = ''

        # Read audio
        audio = self.read_audio(path)

        ## TODO: use decorator to avoid duplicated code.
        try:
            result = self.get_instance().recognize_sphinx(audio)
        except sr.UnknownValueError:
            print(f'Sphinx could not understand {path}')
        except sr.RequestError as error:
            print(f'Sphinx error; {error}')

        debug.trace(7, f'CMUSphinx.speech_to_text({path}) => {result}')
        return result


class Houndify(SpeechRecognition):
    """Houndify Speech Recognition class"""


    # Global state
    _IDENTIFIER = 'houndify'


    def __init__(self, client_id='', client_key=''):
        self.client_id  = client_id
        self.client_key = client_key


    def particular_check_audio(self, path):
        """Check if audio in PATH is valid"""
        is_valid = True

        ## TODO: WORK-IN-PROGRESS
        ## NOTE: Houndify only can be used with audios less than a 1min.

        debug.trace(7, f'SpeechEngine.particular_check_audio({path}) => {is_valid}')
        return is_valid


    def particular_speech_to_text(self, path):
        """Transcribe speech from audio in PATH"""
        result = ''

        # Read audio
        audio = self.read_audio(path)

        # Extract speech
        try:
            result = self.get_instance().recognize_houndify(audio,
                                                            client_id=self.client_id,
                                                            client_key=self.client_key)
        except sr.UnknownValueError:
            print(f'Houndify could not understand {path}')
        except sr.RequestError as error:
            print("Could not request results from Houndify service; {0}".format(error))

        debug.trace(7, f'Houndify.speech_to_text({path}) => {result}')
        return result


class Audio:
    """
    This is a wraper interface for audio processing.
    """


    # Global state
    SPEECH_ENGINES = [CMUSphinx(),
                      Houndify(client_id=HOUNDIFY_ID, client_key=HOUNDIFY_KEY),
                      IBMWatson(api_key=IBM_WATSON_KEY, api_url=IBM_WATSON_URL)]


    def __init__(self, path=''):
        self._path = path


    def set_path(self, path):
        """Set new audio PATH"""
        debug.trace(7, f'Audio.set_path(path={path})')
        self._path = path


    def get_path(self):
        """Get audio PATH"""
        debug.trace(7, f'Audio.get_path() => {self._path}')
        return self._path


    def get_extension(self):
        """Get extension of audio file"""

        result = re.search(r'(?P<path>[\w\-\/]+)\.*(?P<ext>[a-zA-Z0-9]+)* *$', self.get_path())
        result = result.group('ext') if result else ''

        debug.trace(7, f'Audio.get_extension() => {result}')
        return result


    ## pylint: disable=no-self-use
    def get_vocals(self, output=''):
        """Isolate vocals of an audio in path and save on OUTPUT"""

        ## TODO: WORK-IN-PROGRESS

        debug.trace(7, f'SpeechEngine.get_vocals() => result saved on {output}')


    def speech_to_text(self, engine=''):
        """Transcribe speech from audio path using specified ENGINE"""

        # Isolate vocal to improve accuracy
        ## TODO: WORK-IN-PROGRESS

        # Select speech engine.
        speech_engine = self.SPEECH_ENGINES[0]
        if engine:
            for temp_engine in self.SPEECH_ENGINES:
                if temp_engine.get_identifier().lower() == engine.lower():
                    speech_engine = temp_engine
                    break

        # Extract final speech
        result = speech_engine.speech_to_text(self._path)

        debug.trace(7, f'Audio.speech_to_text(path={self._path}, engine={engine}) => {result}')
        return result


if __name__ == '__main__':
    app = AudioArgumentProcessing(description          = __doc__,
                                  positional_arguments = [(SOURCE, 'audio source file')],
                                  text_options         = [(OUTPUT, 'output folder'),
                                                          (SPEECH, 'speech engine')],
                                  boolean_options      = [(VERBOSE, 'verbose print')],
                                  manual_input         = True)
    app.run()



#!/usr/bin/env python3

"""A document object used for the "term frequency" (TF) in TF-IDF.

Example:
    >>> from mezcla.tfidf.document import Document
    >>> from mezcla.tfidf.preprocess import Preprocessor
    >>> input_text = '''Mary had a little lamb, his fur was white as snow
            Everywhere the child went, the lamb, the lamb was sure to go'''
    >>> d = Document(input_text, Preprocessor(stemmer=lambda x: x))
    >>> d__gramsize 
    1
    >>> d.tf_freq('lamb')
    3
    >>> d.tf_raw('lamb')
    0.04411764705882353

"""
# HACK: Attempt at fixing up a very stale example!
# TODO: Make sure the tf_raw calculation revision is correct.
# TODO1: revise docstrings wrt preferred keywords (i.e., not deprecated)

from __future__ import absolute_import, division

# Standard packages
import math
## OLD: import os
import random
## OLD: import sys

# Installed packages
from mezcla import debug
from mezcla import system

# Local packages
## OLD: from .preprocess import clean_text
from mezcla.tfidf.preprocess import clean_text, Preprocessor

# TPO: environment option for weight singleton occurrences low
PENALIZE_SINGLETONS = system.getenv_boolean("PENALIZE_SINGLETONS", False,
                                            "Ignore singleton ngrams")
if PENALIZE_SINGLETONS:
    system.print_stderr("FYI: Penalizing singleton ngrams")


class Document(object):
    """A document holds text, slices text into ngrams, and calculates tf score.

    Attributes:
        text (list): cleaned text, set on init
    """

    def __init__(self, raw_text, preprocessor):
        """All you need is the text body and gramsize (number words in ngram).

        raw_text
            text string input. Will be run through text preprocessing
        preprocessor
            initalized instance of a preprocessor
        """
        self.id = None
        self.text = clean_text(raw_text)
        self.__keywordset = None
        self.__max_raw_frequency = None
        self.__length = None
        self.preprocessor = preprocessor

    def __contains__(self, ngram):
        """Check if the ngram is present in the document."""
        return ngram in self.keywordset

    def __getitem__(self, ngram):
        """Return the DocKeyword object with occurrences via the stemmed ngram."""
        return self.keywordset[ngram]

    def __len__(self):
        """The length of the document is the number of ngrams."""
        # TODO: rename to __size__??
        if not self.__length:
            self.__length = sum([len(x) for x in self.keywordset])
        return self.__length

    @property
    def max_raw_frequency(self):
        """Max ngram frequency found in document."""
        ## if not self.__max_raw_frequency:
        if self.__max_raw_frequency is None:
            biggest_kw = ''
            for kw in self.keywordset:
                if len(kw) > len(biggest_kw):
                    biggest_kw = kw
            self.__max_raw_frequency = len(biggest_kw)
        return self.__max_raw_frequency

    @property
    def gramset(self):
        """Important for fast check if ngram in document.
        Note: Converts self.keywordset hash to set proper.
        """
        ## NOTE: 
        ## OLD: return set([x for x in self.keywordset])
        ## OLD: return {x for x in self.keywordset}
        return set(self.keywordset)

    @property
    def keywordset(self):
        """Return a set of keywords in the document with all their locations."""
        if not self.__keywordset:
            self.__keywordset = {}
            for kw in self.keywords:
                if kw.text not in self.__keywordset:
                    self.__keywordset[kw.text] = kw
                else:
                    self.__keywordset[kw.text] += kw
        return self.__keywordset

    # Term Frequency weighting functions:
    def tf_raw(self, ngram):
        """The (relative) frequency of an ngram in a document."""
        num_occurrences = len(self[ngram]) if ngram in self else 0
        # HACK: give singletons a max DF to lower IDF score
        if (num_occurrences == 1) and PENALIZE_SINGLETONS:
            num_occurrences = 0
        tf_raw = float(num_occurrences) / len(self)
        debug.trace_fmt(7, "tf_raw({ng}): num_occ={no} len(self)={l} result={r}",
                        ng=ngram, no=num_occurrences, l=len(self), r=tf_raw)
        return tf_raw

    def tf_log(self, ngram):
        """The log frequency of an ngram in a document."""
        # TODO: is this formula right? Wikipedia is often wrong...
        return 1 + math.log(self.tf_raw(ngram))

    def tf_binary(self, ngram):
        """Binary term frequency (0 if not present, 1 if ngram is present)."""
        if ngram in self:
            return 1
        return 0

    def tf_norm_50(self, ngram):
        """Double normalized ngram frequency."""
        term_frequency = self.tf_raw(ngram)
        return 0.5 + (0.5 * (term_frequency / self.max_raw_frequency))

    def tf_freq(self, ngram):
        """Returns frequency count for NGRAM"""
        num_occurrences = len(self[ngram]) if ngram in self else 0
        debug.trace_fmt(7, "tf_freq({ng}): num_occ={no} len(self)={l} result={r}",
                        ng=ngram, no=num_occurrences, l=len(self), r=num_occurrences)
        return num_occurrences
    
    def tf(self, ngram, tf_weight='basic', normalize=False):
        """Calculate term frequency.

        Normalizing (stemming) is slow, so it's assumed you're using the ngram.
        """
        if tf_weight == 'norm':
            tf_weight = 'basic'
        if normalize:
            ## TPO: fix referernce to normalize_term
            ## BAD: ngram = self.normalize_term(ngram)
            ngram = self.preprocessor.normalize_term(ngram)
        ## OLD
        ## if tf_weight == 'log':
        ##     return self.tf_log(ngram)
        ## elif tf_weight == 'norm_50':
        ##     return self.tf_norm_50(ngram)
        ## elif tf_weight == 'binary':
        ##     return self.tf_binary(ngram)
        ## elif tf_weight == 'basic':
        ##     return self.tf_raw(ngram)
        ## # TPO: allow for raw frequency
        ## elif tf_weight == 'freq':
        ##     return self.tf_freq(ngram)
        if tf_weight == 'log':
            return self.tf_log(ngram)
        if tf_weight == 'norm_50':
            return self.tf_norm_50(ngram)
        if tf_weight == 'binary':
            return self.tf_binary(ngram)
        if tf_weight == 'basic':
            return self.tf_raw(ngram)
        # TPO: allow for raw frequency
        if tf_weight == 'freq':
            return self.tf_freq(ngram)
        raise ValueError("Invalid tf_weight: " + tf_weight)

    def randgram(self):
        """Return a random gram. Limited to first 100 ngrams."""
        return self.randkeyword().text

    def randkeyword(self):
        """Return a random keyword. Limited to first 100 ngrams."""
        rand_pos = random.randint(0, min(100, len(self.keywordset) - 1))
        for i, x in enumerate(self.keywordset.values()):
            if i == rand_pos:
                return x
        ## print('uh oh')
        ## print(rand_pos)
        system.print_stderr("Error: unexpected condition in randkeyword; rand_pos={rp}", rp=rand_pos)
        return "n/a"

    @property
    def keywords(self):
        """Use the preprocessor to yield keywords from the source text."""
        return self.preprocessor.yield_keywords(self.text, document=self)


def main():
    """Entry point for script: just runs a simple test"""
    text = "my man fran is not a man"
    d = Document(text, Preprocessor(gramsize=1, stemmer=lambda x: x))
    debug.trace_expr(5, [d.tf_freq(t) for t in text.split()])
    debug.assertion(d.tf_freq("man") > d.tf_freq("fran"))
    
#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    system.print_stderr(f"Warning: {__file__} is not intended to be run standalone. A simple test willl be run.")
    main()
